//Copyright (c) 2000-2006 San Diego Supercomputer Center (SDSC),
//a facility operated by the University of California,
//San Diego (UCSD), San Diego, California, USA.
//
//Users and possessors of this source code are hereby granted a
//nonexclusive, royalty-free copyright and design patent license to
//use this code in individual software.  License is not granted for
//commercial resale, in whole or in part, without prior written
//permission from SDSC.  This source is provided "AS IS" without express
//or implied warranty of any kind.
//

package edu.sdsc.sirius.util;

import edu.sdsc.mbt.*;

import java.util.*;


import edu.sdsc.mbt.util.Algebra;
import edu.sdsc.mbt.util.ElementProperties;
import edu.sdsc.mbt.util.Status;
import edu.sdsc.mbt.viewers.GLViewerImpl.*;
import edu.sdsc.sirius.viewers.*;

/**
 * This class contains all the math necessary to transform atomic coordinates and
 * drive rotation and translation when building structures
 * @author Oleksandr V. Buzko
 *
 */
public class StructureMath {
	
	private static HashMap pka = ProteinProperties.getPKATable();
	private static HashMap dna = DNAProperties.getPKATable();
	
	private static double[] transform = new double[16];
	private static double[] vectorFromOrigin = new double[3];
	
	public static final void transformStructure(Location start, Location target, Structure structure){
				
		Algebra.normalizeVector(start.lookAt);
		Algebra.normalizeVector(start.up);
		Algebra.normalizeVector(target.lookAt);
		Algebra.normalizeVector(target.up);
		
		//first, check if the input locations are the same
		boolean lv = start.equalLookAtVectors(target);
		boolean uv = start.equalUpVectors(target);
		boolean p = start.equalPositions(target);
		
//		System.out.println("start = " + start);
//		System.out.println("target = " + target);
		
//		System.out.println("lv = " + lv + ", uv = " + uv + ", p = " + p);
		
		//check if the vectors are the same
		if (lv && uv && p) return;
		
//		System.out.println("Transforming");
		
		//calculate the displacement in each coordinate between the target location and the origin
		double dx2 = target.position[0];
		double dy2 = target.position[1];
		double dz2 = target.position[2];
		
		//calculate offsets to reset coordinates to the origin
		double dx = -start.position[0];
		double dy = -start.position[1];
		double dz = -start.position[2];
		
		
//		System.out.println("dx = " + dx + ", dx2 = " + dx2);
		
		if (!lv){
		
			//get the matrix to apply rotation
			double[][] matrix = getRotationMatrix(start.lookAt, target.lookAt);
	
			//apply the change to coordinates of structure
			StructureMap structureMap = structure.getStructureMap();
			for (int i = 0; i < structureMap.getAtomCount(); i++){
				Atom a = structureMap.getAtom(i);
//				System.out.println("atom before = " + a.coordinate[0] + "," + a.coordinate[1] + "," + a.coordinate[2]);
				a.coordinate[0] += dx;
				a.coordinate[1] += dy;
				a.coordinate[2] += dz;
				
				transformCoordinates(a.coordinate, matrix);
//				System.out.println("atom after = " + a.coordinate[0] + "," + a.coordinate[1] + "," + a.coordinate[2]);
			}
			
			transformCoordinates(start.up, matrix);
		}
		
		//check if the new up vector is the same as the target up vector
		boolean up = (start.up[0] == target.up[0] && start.up[1] == target.up[1] && start.up[2] == target.up[2]);
		
		
//		System.out.println("start.up before = " + start.up[0] + "," + start.up[1] + "," + start.up[2]);
		
		
		//get the up vector rotation matrix
		//check is start.up and target.up are the same
//		System.out.println("start.up = " + start.up[0] + "," + start.up[1] + "," + start.up[2]);
//		System.out.println("target.up = " + target.up[0] + "," + target.up[1] + "," + target.up[2]);
		
		double[][] matrix2 = getRotationMatrix(start.up, target.up);
		
		for (int i = 0; i < structure.getStructureMap().getAtomCount(); i++){
			Atom a = structure.getStructureMap().getAtom(i);
			
			if (!up){
				transformCoordinates(a.coordinate, matrix2);
			}
			
			//return the molecule to its position in space
			a.coordinate[0] += dx2;
			a.coordinate[1] += dy2;
			a.coordinate[2] += dz2;
		
//			System.out.println("atom after = " + a.coordinate[0] + "," + a.coordinate[1] + "," + a.coordinate[2]);
		}
		
	}
	
	public static final double[] transformPoint(Location start, Location target, double[] point){
		
		double[] out = new double[3];
		out[0] = point[0];
		out[1] = point[1];
		out[2] = point[2];
		
		Algebra.normalizeVector(start.lookAt);
		Algebra.normalizeVector(start.up);
		Algebra.normalizeVector(target.lookAt);
		Algebra.normalizeVector(target.up);
		
		//first, check if the input locations are the same
		boolean lv = start.equalLookAtVectors(target);
		boolean uv = start.equalUpVectors(target);
		boolean p = start.equalPositions(target);
		
		if (lv && uv && p){
			return out;
		}
		
		//calculate the displacement in each coordinate between the target location and the origin
		double dx2 = target.position[0];
		double dy2 = target.position[1];
		double dz2 = target.position[2];
		
		//calculate offsets to reset coordinates to the origin
		double dx = -start.position[0];
		double dy = -start.position[1];
		double dz = -start.position[2];
		
		
		if (!lv){
		
			//get the matrix to apply rotation
			double[][] matrix = getRotationMatrix(start.lookAt, target.lookAt);
	
			out[0] += dx;
			out[1] += dy;
			out[2] += dz;
				
			transformCoordinates(out, matrix);
			
			transformCoordinates(start.up, matrix);
		}
		
		//check if the new up vector is the same as the target up vector
		boolean up = (start.up[0] == target.up[0] && start.up[1] == target.up[1] && start.up[2] == target.up[2]);
		
		
		if (!up){
			double[][] matrix2 = getRotationMatrix(start.up, target.up);
			transformCoordinates(out, matrix2);
		}
		
		//relocate the point to its position in space
		out[0] += dx2;
		out[1] += dy2;
		out[2] += dz2;
		
		return out;
		
	}
	
	public static final void transformAtomSet(Location start, Location target, Vector atoms){
		
		Algebra.normalizeVector(start.lookAt);
		Algebra.normalizeVector(start.up);
		Algebra.normalizeVector(target.lookAt);
		Algebra.normalizeVector(target.up);
		
		//first, check if the input locations are the same
		boolean lv = start.equalLookAtVectors(target);
		boolean uv = start.equalUpVectors(target);
		boolean p = start.equalPositions(target);
		
		if (lv && uv && p) return;
		
		//calculate the displacement in each coordinate between the target location and the origin
		double dx2 = target.position[0];
		double dy2 = target.position[1];
		double dz2 = target.position[2];
		
		//calculate offsets to reset coordinates to the origin
		double dx = -start.position[0];
		double dy = -start.position[1];
		double dz = -start.position[2];
		
		
		if (!lv){
		
			//get the matrix to apply rotation
			double[][] matrix = getRotationMatrix(start.lookAt, target.lookAt);
	
			//apply the change to coordinates of structure
			for (int i = 0; i < atoms.size(); i++){
				Atom a = (Atom)atoms.get(i);
				a.coordinate[0] += dx;
				a.coordinate[1] += dy;
				a.coordinate[2] += dz;
				
				transformCoordinates(a.coordinate, matrix);
			}
			
			transformCoordinates(start.up, matrix);
		}
		
		//check if the new up vector is the same as the target up vector
		boolean up = (start.up[0] == target.up[0] && start.up[1] == target.up[1] && start.up[2] == target.up[2]);
		
		double[][] matrix2 = getRotationMatrix(start.up, target.up);
		
		for (int i = 0; i < atoms.size(); i++){
			Atom a = (Atom)atoms.get(i);
			
			if (!up){
				transformCoordinates(a.coordinate, matrix2);
			}
			
			//return the atom to its position in space
			a.coordinate[0] += dx2;
			a.coordinate[1] += dy2;
			a.coordinate[2] += dz2;
		
		}
	}
	
	public static final void transformVector(Location start, Location target, double[] vector, double[] out){
		
		Algebra.normalizeVector(start.lookAt);
		Algebra.normalizeVector(start.up);
		Algebra.normalizeVector(target.lookAt);
		Algebra.normalizeVector(target.up);
		
		//first, check if the input locations are the same
		boolean lv = start.equalLookAtVectors(target);
		boolean uv = start.equalUpVectors(target);
		boolean p = start.equalPositions(target);
		
		out[0] = vector[0];
		out[1] = vector[1];
		out[2] = vector[2];
		
		if (lv && uv && p) return;
		
		//calculate the displacement in each coordinate between the target location and the origin
		double dx2 = target.position[0];
		double dy2 = target.position[1];
		double dz2 = target.position[2];
		
		//calculate offsets to reset coordinates to the origin
		double dx = -start.position[0];
		double dy = -start.position[1];
		double dz = -start.position[2];
		
		
		if (!lv){
		
			//get the matrix to apply rotation
			double[][] matrix = getRotationMatrix(start.lookAt, target.lookAt);
	
			//apply the change to coordinates of structure
			out[0] += dx;
			out[1] += dy;
			out[2] += dz;
			
			transformCoordinates(out, matrix);
			
			transformCoordinates(start.up, matrix);
		}
		
		//check if the new up vector is the same as the target up vector
		boolean up = (start.up[0] == target.up[0] && start.up[1] == target.up[1] && start.up[2] == target.up[2]);
		
		double[][] matrix2 = getRotationMatrix(start.up, target.up);
		
			
		if (!up){
			transformCoordinates(out, matrix2);
		}
		
		//return the atom to its position in space
		out[0] += dx2;
		out[1] += dy2;
		out[2] += dz2;
		
		
	}
	
	public static final Vector transformVectors(double[] start, double[] target, Vector vectors){
		
		Algebra.normalizeVector(start);
		Algebra.normalizeVector(target);
		
		Vector out = new Vector();
		for (int i = 0; i < vectors.size(); i++){
			double[] a = new double[]{ ((double[])vectors.get(i))[0], ((double[])vectors.get(i))[1], ((double[])vectors.get(i))[2] };
			out.add(a);
		}
		
		if (start[0] == target[0] && start[1] == target[1] && start[2] == target[1]){
			return out;
		}
		
		//get the matrix to apply rotation
		double[][] matrix = getRotationMatrix(start, target);

		//apply the change
		for (int i = 0; i < out.size(); i++){
			double[] a = (double[])out.get(i);
			transformCoordinates(a, matrix);
		
		}
			
		return out;
		
	}


	
	/**
	 * This method returns the new x,y,z coordinates for a point given the corresponding
	 * rotation matrix
	 * @param input
	 * @param matrix
	 * @return
	 */
	public static final void transformCoordinates(double[] input, double[][] matrix){
		
		double a = matrix[0][0]*input[0] + matrix[0][1]*input[1] + matrix[0][2]*input[2];
		double b = matrix[1][0]*input[0] + matrix[1][1]*input[1] + matrix[1][2]*input[2];
		double c = matrix[2][0]*input[0] + matrix[2][1]*input[1] + matrix[2][2]*input[2];
		
		input[0] = a;
		input[1] = b;
		input[2] = c;
	}
	
	public static final void transformCoordinates(double[] input, double[] matrix){
		
		if (matrix.length != 16) return;
		
		double a = matrix[0]*input[0] + matrix[1]*input[1] + matrix[2]*input[2] + matrix[12]*1;
		double b = matrix[4]*input[0] + matrix[5]*input[1] + matrix[6]*input[2] + matrix[13]*1;
		double c = matrix[8]*input[0] + matrix[9]*input[1] + matrix[10]*input[2] + matrix[14]*1;
		
		input[0] = a;
		input[1] = b;
		input[2] = c;
	}
	
	/**
	 * Get coordinates of a point on a vector that originates from a given point
	 * @param eye - point fromm where the vector starts
	 * @param vector - vector in question
	 * @param distance
	 * @return
	 */
	public static final double[] getPointOnVector(double[] eye, double[] vector, double distance){
		double[] point = new double[3];
		
		//change vector length
		double n = Math.sqrt(vector[0]*vector[0] + vector[1]*vector[1] + vector[2]*vector[2]);
		double factor = distance/n;
		double x = factor * vector[0];
		double y = factor * vector[1];
		double z = factor * vector[2];
		
		point[0] = x + eye[0];
		point[1] = y + eye[1];
		point[2] = z + eye[2];
				
		return point;
	}
	
	/**
	 * Returns a location of the bond, with the lookAt vector pointed along the bond
	 * and the up vector being orthogonal to the plane of the bond and one of the atoms
	 * the third vector is computed automatically
	 * The position is at the heavy atom
	 * @param bond
	 * @return
	 */
	public static final Location getBondLocation(Bond bond){
		
		Atom a0 = bond.getAtom(0);
		Atom a1 = bond.getAtom(1);
		
		double[] lookAt = new double[3];
		double[] up = new double[3];
		double[] position = new double[3];
		
		double[] temp = new double[3];//vector along an adjacent bond to be used to calculate an orthogonal vector
		
		if (a0.element == 1){
			lookAt[0] = a0.coordinate[0] - a1.coordinate[0];
			lookAt[1] = a0.coordinate[1] - a1.coordinate[1];
			lookAt[2] = a0.coordinate[2] - a1.coordinate[2];
			
			Vector bonds = a1.structure.getStructureMap().getBonds(a1);
			for (int i = 0; i < bonds.size(); i++){
				Bond bb = (Bond)bonds.get(i);
				if (bb != bond){
					//we can use this bond as the other vector
					if (bb.getAtom(0) == a1){
						temp[0] = bb.getAtom(1).coordinate[0] - a1.coordinate[0];
						temp[1] = bb.getAtom(1).coordinate[1] - a1.coordinate[1];
						temp[2] = bb.getAtom(1).coordinate[2] - a1.coordinate[2];
						break;
					}
					else if (bb.getAtom(1) == a1){
						temp[0] = bb.getAtom(0).coordinate[0] - a1.coordinate[0];
						temp[1] = bb.getAtom(0).coordinate[1] - a1.coordinate[1];
						temp[2] = bb.getAtom(0).coordinate[2] - a1.coordinate[2];
						break;
					}
				}
			}
			
			position[0] = a1.coordinate[0];
			position[1] = a1.coordinate[1];
			position[2] = a1.coordinate[2];
			
		}
		else if (a1.element == 1){
			lookAt[0] = a1.coordinate[0] - a0.coordinate[0];
			lookAt[1] = a1.coordinate[1] - a0.coordinate[1];
			lookAt[2] = a1.coordinate[2] - a0.coordinate[2];
			
			Vector bonds = a0.structure.getStructureMap().getBonds(a0);
			for (int i = 0; i < bonds.size(); i++){
				Bond bb = (Bond)bonds.get(i);
				if (bb != bond){
					//we can use this bond as the other vector
					if (bb.getAtom(0) == a0){
						temp[0] = bb.getAtom(1).coordinate[0] - a0.coordinate[0];
						temp[1] = bb.getAtom(1).coordinate[1] - a0.coordinate[1];
						temp[2] = bb.getAtom(1).coordinate[2] - a0.coordinate[2];
						break;
					}
					else if (bb.getAtom(1) == a0){
						temp[0] = bb.getAtom(0).coordinate[0] - a0.coordinate[0];
						temp[1] = bb.getAtom(0).coordinate[1] - a0.coordinate[1];
						temp[2] = bb.getAtom(0).coordinate[2] - a0.coordinate[2];
						break;
					}
					
				}
			}
			position[0] = a0.coordinate[0];
			position[1] = a0.coordinate[1];
			position[2] = a0.coordinate[2];
			
		}
		
		up = Algebra.crossProduct(lookAt, temp);
		
		return new Location(lookAt, up, position);
	}
	
	public static final double[][] getRotationMatrix(double[] vector1, double[] vector2){
		double[][] matrix = new double[3][3];
		
		Algebra.normalizeVector(vector1);
		Algebra.normalizeVector(vector2);
		
		//get the angle between the vectors
		double angle = Math.acos(Algebra.dotProduct(vector1, vector2)/(Algebra.vectorLength(vector1)*Algebra.vectorLength(vector2)));
//		System.out.println("angle = " + angle);
		//if the angle is PI, axis will be undefined, so use any orthogonal vector as the axis
		//for opposite vectors it'll work just fine
		
		//get the axis of rotation first
		double[] axis = Algebra.crossProduct(vector1, vector2);
		if (axis[0] == 0.0 && axis[1] == 0.0 && axis[2] == 0.0){
			//get an axis
			double[] vector3 = new double[]{ vector1[0], vector1[1]+0.1, vector1[2]+0.1 };
			axis = Algebra.crossProduct(vector1, vector3);
		}
		else{
			
		}
		
		//get the auxiliary variables
		double sin = Math.sin(angle);
		double cos = Math.cos(angle);
		double c = 1 - cos;
		
		double m = Algebra.vectorLength(axis);
		double mm = m*m;
		
		
		
		//do the calculation
		matrix[0][0] = (axis[0]*axis[0] + (axis[1]*axis[1] + axis[2]*axis[2])*cos)/mm;
		matrix[0][1] = (axis[0]*axis[1]*c - axis[2]*m*sin)/mm;
		matrix[0][2] = (axis[0]*axis[2]*c + axis[1]*m*sin)/mm;
		matrix[1][0] = (axis[0]*axis[1]*c + axis[2]*m*sin)/mm;
		matrix[1][1] = (axis[1]*axis[1] + (axis[0]*axis[0] + axis[2]*axis[2])*cos)/mm;
		matrix[1][2] = (axis[1]*axis[2]*c - axis[0]*m*sin)/mm;
		matrix[2][0] = (axis[0]*axis[2]*c - axis[1]*m*sin)/mm;
		matrix[2][1] = (axis[1]*axis[2]*c + axis[0]*m*sin)/mm;
		matrix[2][2] = (axis[2]*axis[2] + (axis[0]*axis[0] + axis[1]*axis[1])*cos)/mm;
		
		return matrix;
	}
	
	public static final double[][] getRotationMatrix(double angle, double[] axis){
		double[][] matrix = new double[3][3];
		
		//convert angle to radians
		double radians = (Math.PI * angle)/180.0;
		
//		System.out.println("angle = " + angle + "; radians = " + radians);

		//get the auxiliary variables
		double sin = Math.sin(radians);
		double cos = Math.cos(radians);
		double c = 1 - cos;
		
		double m = Algebra.vectorLength(axis);
		double mm = m*m;
		
		//do the calculation
		matrix[0][0] = (axis[0]*axis[0] + (axis[1]*axis[1] + axis[2]*axis[2])*cos)/mm;
		matrix[0][1] = (axis[0]*axis[1]*c - axis[2]*m*sin)/mm;
		matrix[0][2] = (axis[0]*axis[2]*c + axis[1]*m*sin)/mm;
		matrix[1][0] = (axis[0]*axis[1]*c + axis[2]*m*sin)/mm;
		matrix[1][1] = (axis[1]*axis[1] + (axis[0]*axis[0] + axis[2]*axis[2])*cos)/mm;
		matrix[1][2] = (axis[1]*axis[2]*c - axis[0]*m*sin)/mm;
		matrix[2][0] = (axis[0]*axis[2]*c - axis[1]*m*sin)/mm;
		matrix[2][1] = (axis[1]*axis[2]*c + axis[0]*m*sin)/mm;
		matrix[2][2] = (axis[2]*axis[2] + (axis[0]*axis[0] + axis[1]*axis[1])*cos)/mm;
		
		return matrix;
		
	}
	
	public static final double[] getLinearRotationMatrix(double angle, double[] axis){
		
		double[] out = new double[16];//linear output matrix
		
		double[][] matrix = getRotationMatrix(-angle, axis);
		
		out[0] = matrix[0][0];
		out[1] = matrix[0][1];
		out[2] = matrix[0][2];
		out[3] = 0.0;
		out[4] = matrix[1][0];
		out[5] = matrix[1][1];
		out[6] = matrix[1][2];
		out[7] = 0.0;
		out[8] = matrix[2][0];
		out[9] = matrix[2][1];
		out[10] = matrix[2][2];
		out[11] = 0.0;
		out[12] = 0.0;
		out[13] = 0.0;
		out[14] = 0.0;
		out[15] = 1.0;
		
		return out;
		
	}
	
	public static final double[] getLinearTranslationMatrix(double[] vector){
		double[] out = new double[16];
		out[0] = 1.0;
		out[1] = 0.0;
		out[2] = 0.0;
		out[3] = 0.0;
		out[4] = 0.0;
		out[5] = 1.0;
		out[6] = 0.0;
		out[7] = 0.0;
		out[8] = 0.0;
		out[9] = 0.0;
		out[10] = 1.0;
		out[11] = 0.0;
		out[12] = vector[0];
		out[13] = vector[1];
		out[14] = vector[2];
		out[15] = 1.0;

		
		return out;
	}
	
	/**
	 * Note that the argument is in degrees, and they get converted into radians
	 * @param axis
	 * @param root
	 * @param angle
	 * @param atoms
	 */
	public static final void rotateStructure(double[] axis, Atom root, double angle, Vector atoms){
		
		if (angle == 0 || atoms == null || atoms.size() == 0) return;
		if (axis == null || Algebra.vectorLength(axis) == 0) return;
		
		double[][] matrix = getRotationMatrix(angle, axis);
		
/*		System.out.println("matrix[0] = " + matrix[0][0]);
		System.out.println("matrix[1] = " + matrix[0][1]);
		System.out.println("matrix[2] = " + matrix[0][2]);
		System.out.println("matrix[3] = " + matrix[1][0]);
		System.out.println("matrix[4] = " + matrix[1][1]);
		System.out.println("matrix[5] = " + matrix[1][2]);
		System.out.println("matrix[6] = " + matrix[2][0]);
		System.out.println("matrix[7] = " + matrix[2][1]);
		System.out.println("matrix[8] = " + matrix[2][2]);
*/
		
		double dx = root.coordinate[0];
		double dy = root.coordinate[1];
		double dz = root.coordinate[2];
		
		//apply the change to coordinates of structure
		for (int i = 0; i < atoms.size(); i++){
			Atom a = (Atom)atoms.get(i);
			//move the atom to the origin
			
			a.coordinate[0] -= dx;
			a.coordinate[1] -= dy;
			a.coordinate[2] -= dz;
			
			transformCoordinates(a.coordinate, matrix);

			//move atom to the original position
			a.coordinate[0] += dx;
			a.coordinate[1] += dy;
			a.coordinate[2] += dz;
		}
	}
	
	public static final void rotateCoordinates(double[] axis, double[] root, double angle, Vector coordinates){
		
		if (angle == 0 || coordinates == null || coordinates.size() == 0) return;
		if (axis == null || Algebra.vectorLength(axis) == 0) return;
		
		double[][] matrix = getRotationMatrix(angle, axis);
		
		double dx = root[0];
		double dy = root[1];
		double dz = root[2];
		
		//apply the change to coordinates of structure
		for (int i = 0; i < coordinates.size(); i++){
			double[] a = (double[])coordinates.get(i);
			//move the atom to the origin
			
			a[0] -= dx;
			a[1] -= dy;
			a[2] -= dz;
			
			transformCoordinates(a, matrix);

			//move atom to the original position
			a[0] += dx;
			a[1] += dy;
			a[2] += dz;
		}
	}

	
	public static final double[] getToOriginMatrix(Atom root){
		return StructureMath.getToOriginMatrix(root.coordinate);
	}
	
	public static final double[] getToOriginMatrix(double[] coordinates){
		double[] out = new double[16];
		
		double dx = coordinates[0];
		double dy = coordinates[1];
		double dz = coordinates[2];
		
		out[0] = 1.0;
		out[1] = 0.0;
		out[2] = 0.0;
		out[3] = 0.0;
		out[4] = 0.0;
		out[5] = 1.0;
		out[6] = 0.0;
		out[7] = 0.0;
		out[8] = 0.0;
		out[9] = 0.0;
		out[10] = 1.0;
		out[11] = 0.0;
		out[12] = -dx;
		out[13] = -dy;
		out[14] = -dz;
		out[15] = 1.0;
		
		return out;
		
	}
	
	public static final double[] getFromOriginMatrix(Atom root){
		return StructureMath.getFromOriginMatrix(root.coordinate);
	}
	
	public static final double[] getFromOriginMatrix(double[] coordinates){
		double[] out = new double[16];
		
		double dx = coordinates[0];
		double dy = coordinates[1];
		double dz = coordinates[2];
		
		out[0] = 1.0;
		out[1] = 0.0;
		out[2] = 0.0;
		out[3] = 0.0;
		out[4] = 0.0;
		out[5] = 1.0;
		out[6] = 0.0;
		out[7] = 0.0;
		out[8] = 0.0;
		out[9] = 0.0;
		out[10] = 1.0;
		out[11] = 0.0;
		out[12] = dx;
		out[13] = dy;
		out[14] = dz;
		out[15] = 1.0;
		return out;
		
	}
	
	/**
	 * This method runs through the structure and adds hydrogens where necessary, based on
	 * tabulated element valencies. Things like incorrect bond orders are ignored, so it's the
	 * user's responsibility (for now) to take care of the correct bond orders.
	 * If the number of hydrogens is too great, the number is reduced to the valency and geometry
	 * is recalculated
	 * @param structure
	 */
	public static void checkHydrogens(Structure structure, StructureViewer viewer, float pH){
		StructureMap map = structure.getStructureMap();
				
		HashMap addedAtoms = new HashMap();
		
		boolean progress = false;
		float percentDone = 1.0f;
		int onePercent = 1;
		
//		System.out.println("atom count = " + structure.getStructureMap().getAtomCount());
		if (structure.getStructureMap().getAtomCount() > 1500){
			progress = true;
			onePercent = ( structure.getStructureMap().getAtomCount()/ 100);
			if ( onePercent <= 0 ) onePercent = 1;
			percentDone = 0.0f;
			Status.progress( percentDone, "Calcul des hydrognes");
		}
		

		for (int i = 0; i < map.getAtomCount(); i++){
			Atom atom = map.getAtom(i);
			if ( progress && i % onePercent == 0 )
			{
				percentDone = (float) i / (float) map.getAtomCount();
				Status.progress( percentDone, "Calcul des hydrognes" );
			}
			checkHydrogens(atom, addedAtoms, viewer, pH);
		}
	
		// Progress is done.
		if (progress)Status.progress( 1.0f, null );
		
		//for this wholesale addition of hydrogens, check available 
		//gaps in atom numbering beforehand, and then use them for 
		//atom addition. Once the gaps are done, start adding numbers
		//on the top of the current atom count
		int[] numbers = new int[map.getAtomCount()];
		for (int i = 0; i < map.getAtomCount(); i++){
			numbers[i] = map.getAtom(i).number;
		}
		
		Vector anumbers = new Vector();//numbers of atoms available in gaps
		
		Arrays.sort(numbers);
		boolean gap = false;
		int last = -1;
		for (int i = 0; i < numbers.length; i++){
			if (last == -1){
				last = numbers[i];
				continue;
			}
			int dif = numbers[i] - last;
			if (dif > 1){
				//this is a gap
				anumbers.add(new Integer(numbers[i] - 1));
			}
			last = numbers[i];
		}
		
		int start = map.getAtomCount() + 1;
		for (int i = 0; i < addedAtoms.size(); i++){
			if (i >= anumbers.size()){
				anumbers.add(new Integer(start));
				start++;
			}
		}
		
		//in the end, anumbers contains all numbers that should be assigned in order
		if (progress){
			onePercent = ( addedAtoms.size()/ 100);
			if ( onePercent <= 0 ) onePercent = 1;
			percentDone = 0.0f;
			Status.progress( percentDone, "Mise  jour de la structure");
		}
		Set keys = addedAtoms.keySet();
		Iterator it = keys.iterator();
		int counter = 0;
		while (it.hasNext()){
			Atom a = (Atom)it.next();
			Atom b = (Atom)addedAtoms.get(a);
			Bond bb = new Bond(a, b);
			bb.setOrderOnly(1.0f);
			
			map.addAtom(a, b, bb, ((Integer)anumbers.get(counter)).intValue());
			
			//create a geometry for the new bond
			viewer.addAtom(a, false);
			if ( progress && counter % onePercent == 0 )
			{
				percentDone = (float) counter / (float) addedAtoms.size();
				Status.progress( percentDone, "Mise  jour de la structure" );
			}
			counter++;
		}
		
		// Progress is done.
		if (progress) Status.progress( 1.0f, null );
	
		
		viewer.updateAppearance();
	}
	
	public static void checkHydrogens(Atom atom, StructureViewer viewer, float pH, boolean batch){
		double bondLength = 1.2;//later take it from a bond length table
		
		HashMap addedAtoms = new HashMap();
		
		checkHydrogens(atom, addedAtoms, viewer, pH);
		
//		System.out.println("Added atoms = " + addedAtoms.size());
		Set keys = addedAtoms.keySet();
		Iterator it = keys.iterator();
		while (it.hasNext()){
			Atom a = (Atom)it.next();
			Atom b = (Atom)addedAtoms.get(a);
			Bond bb = new Bond(a, b);
			bb.setOrderOnly(1.0f);
			if (ProteinProperties.isAminoAcid(atom.compound) || DNAProperties.isNucleicAcid(atom.compound)){
				atom.structure.getStructureMap().addAtom(a, b, bb, false);
			}
			else{
				atom.structure.getStructureMap().addAtom(a, b, bb, true);
			}
			
			
			//create a geometry for the new bond
			if (viewer != null) viewer.addAtom(a, batch);
		}
		
	}
	
	private static void checkHydrogens(Atom atom, HashMap addedAtoms, StructureViewer viewer, float pH){
		
		if (atom.element == 7){
			//check if this N has a carbonyl next to it
			Vector next = atom.structure.getStructureMap().getAtoms(atom);
			if (next == null || next.size() == 0) return;
			
			for (int j = 0; j < next.size(); j++){
				Atom aa = (Atom)next.get(j);
				if (aa.element == 6){
					//it's a carbon. check if it has an oxygen
					//connected by a double bond
					Vector bonds = atom.structure.getStructureMap().getBonds(aa);
					for (int k = 0; k < bonds.size(); k++){
						Bond bond = (Bond)bonds.get(k);
						if (bond.order == 2){
							//check if the other atom is an oxygen
							if (bond.getAtom(0) == aa){
								if (bond.getAtom(1).element == 8){
									atom.hybridization = Atom.SP2;
									break;
								}
							}
							else if (bond.getAtom(1) == aa){
								if (bond.getAtom(0).element == 8){
									atom.hybridization = Atom.SP2;
									break;
								}
							}
						}
					}
				}
			}
		}

		
		StructureMap map = atom.structure.getStructureMap();
		Vector bonds = map.getBonds(atom);
//		System.out.println("bonds size = " + bonds.size() + " for " + atom.name + " in " + atom.residue.getCompoundCode());
		int[] v = ElementProperties.getElementValency(atom.element);
		if (v == null) v = new int[]{ 1 };
		//simplest case first
		//get the total bond count, checking for double bonds
		float total = 0.0f;
		if (bonds == null || bonds.size() == 0){
			return;
		}
		for (int j = 0; j < bonds.size(); j++){
			Bond b = (Bond)bonds.get(j);
//			System.out.println("bond = " + b.getAtom(0).name + " - " + b.getAtom(1).name + ": " + b.order);
			total += ((Bond)bonds.get(j)).order;
		}
//		System.out.println("total at atom " + atom.name + " is " + total);
		int valency = -1;//default to make sure we can exit early if no hydrogen addition is needed
		if (v.length == 1){
			valency = v[0];
//			System.out.println("valency set to " + valency);
		}
		else if (atom.hybridization > Atom.SP3 && v.length > 1){//let's assume the lowest valency to be effective in multiple bonds
			valency = v[0];
		}
		else{
			//get the effective valency for this case
			for (int k = 0; k < v.length; k++){
				if (v[k] == total){
					//valencies are satisfied
					valency = v[k];
					break;
				}
				else if (v[k] > total){
					valency = v[k];
					break;
				}
			}
		}
		if (valency == -1){
			return;
		}
		
		//take into account formal charges
		valency += atom.formalCharge;
		
//		System.out.println("valency = " + valency + ", total = " + total + " for " + atom.name);
		
		if (total < (float)valency){
			
			
			//get the difference and add missing hydrogens
			float dif = (float)valency - total;
//			System.out.println("dif = " + dif);
			if (dif < 1.0) return;//don't do anything if the difference is too small (like in partial bond order changes)
			
			//check if there are already hydrogens around this atom.
			//if yes, delete them, since geometry may have changed, and add them again in a new pass
			Vector aas = atom.structure.getStructureMap().getAtoms(atom);
			Vector doomed = new Vector();
			for (int k = 0; k < aas.size(); k++){
				if (((Atom)aas.get(k)).element == 1){
					doomed.add(aas.get(k));
				}
			}
			
			if (doomed.size() > 0){
				for (int k = 0; k < doomed.size(); k++){
					if (viewer != null) viewer.deleteHydrogen((Atom)doomed.get(k), false);
				}
				checkHydrogens(atom, addedAtoms, viewer, pH);
				return;
			}
			
			//play with the independent motion option
			HashMap transformedCoordinates = viewer.getTransformedCoordinates();
			
			Bond b = (Bond)bonds.get(0);
			Atom next = null;
			if (b.getAtom(0) == atom){
				next = b.getAtom(1);
			}
			else if (b.getAtom(1) == atom){
				next = b.getAtom(0);
			}
			
			double[] atomCoordinates = null;
			double[] nextCoordinates = null;
			
			if (transformedCoordinates != null && transformedCoordinates.containsKey(atom)){
				atomCoordinates = (double[])transformedCoordinates.get(atom);
			}
			else{
				atomCoordinates = atom.coordinate;
			}
			
			if (transformedCoordinates != null && transformedCoordinates.containsKey(next)){
				nextCoordinates = (double[])transformedCoordinates.get(next);
			}
			else{
				nextCoordinates = next.coordinate;
			}

			
			double[] direction = new double[]{ atomCoordinates[0] - nextCoordinates[0], atomCoordinates[1] - nextCoordinates[1], atomCoordinates[2] - nextCoordinates[2]}; 


			//get the vector of the bond
			double[] bond = new double[]{ nextCoordinates[0] - atomCoordinates[0], nextCoordinates[1] - atomCoordinates[1],
					nextCoordinates[2] - atomCoordinates[2] };

//			System.out.println("atom:" + atom.name + ": " + atom.coordinate[0] + ", " + atom.coordinate[1] + "," + atom.coordinate[2]);
//			System.out.println("next:" + next.name + ": " + next.coordinate[0] + ", " + next.coordinate[1] + "," + next.coordinate[2]);

			double[] normal = new double[3];
			
			int factor = 2;//offset of vector indices

			//calculate the orientation normal depending on the case: terminal bond vs in-chain
			if (bonds.size() == 1){//only one bond (single or double)
				
				//get the normal to the plane of the new bond(s)

				Vector atoms = map.getAtoms(next);
				if (atoms.size() == 1){//only the parent atom is present
					//this is just a short molecule, so add the missing atoms in any orientation
//					System.out.println("short at " + atom.name);
					//get some random normal
					//create a similar vector and get their cross-product
					double[] second = new double[]{ direction[0], direction[1], direction[2]+0.1};
					normal = Algebra.crossProduct(direction, second);
				}
				else{
					Atom a = (Atom)atoms.get(0);
					if (a == atom){
						a = (Atom)atoms.get(1);
					}
					//get the normal vector
					if (transformedCoordinates != null && transformedCoordinates.containsKey(a)){
						normal = BondGeometryUtils.getNormalVector(atomCoordinates, nextCoordinates, (double[])transformedCoordinates.get(a));
					}
					else{
						normal = BondGeometryUtils.getNormalVector(atom.coordinate, next.coordinate, a.coordinate);
					}
				}
			}
			else{
				Bond b2 = (Bond)bonds.get(1);
				//get the total list of atoms around atom, and later check which one has not been
				//used for normal calculation if the total number is 3
				Atom next2 = null;
				if (b2.getAtom(0) == atom){
					next2 = b2.getAtom(1);
				}
				else if (b2.getAtom(1) == atom){
					next2 = b2.getAtom(0);
				}
				
				if (transformedCoordinates != null && transformedCoordinates.containsKey(next2)){
					normal = BondGeometryUtils.getNormalVector(nextCoordinates, atomCoordinates, (double[])transformedCoordinates.get(next2));
				}
				else{
					normal = BondGeometryUtils.getNormalVector(next.coordinate, atom.coordinate, next2.coordinate);
				}
//				System.out.println("atom = " + atom.name + ", next = " + next.name + ", next2 = " + next2.name);
//				System.out.println("bond = " + bond[0] + "," + bond[1] + "," + bond[2]);
//				double[] nextBond = new double[]{ next2.coordinate[0] - atom.coordinate[0], next2.coordinate[1] - atom.coordinate[1], next2.coordinate[2] - atom.coordinate[2] };
				
				
				
//				double[] end1 = getPointOnVector(atom.coordinate, normal, 1.0);
//				viewer.createArrow(atom.structure, atom.coordinate, end1, new float[]{ 0.0f, 0.5f, 1.0f});

				factor = 3;
				
				if (bonds.size() == 3){
					Vector atoms = map.getAtoms(atom);
					for (int m = 0; m < atoms.size(); m++){
						if (atoms.get(m) != next && atoms.get(m) != next2){
							//it's the third atom and should be used as the reference
							Atom aaa = (Atom)atoms.get(m);
//							System.out.println("aaa = " + aaa.name);
							double angle = 0;
							if (transformedCoordinates != null && transformedCoordinates.containsKey(aaa)){
								angle = BondGeometryUtils.getVectorAngle(normal, atomCoordinates, (double[])transformedCoordinates.get(aaa));
							}
							else{
								angle = BondGeometryUtils.getVectorAngle(normal, atom.coordinate, aaa.coordinate);
							}
//							System.out.println("angle at " + atom.name + " is " + angle);
							//check the angle between the normal and this vector
							if (angle < Math.PI/2){
								factor = 4;
							}
							break;
						}
					}
				}

			}

//			System.out.println("normal = " + normal[0] + "," + normal[1] + "," + normal[2]);
										
//			double[] end1 = getPointOnVector(atom.coordinate, normal, 1.0);
//			viewer.createArrow(structure, atom.coordinate, end1, new float[]{ 1.0f, 1.0f, 0.0f});
			
//			System.out.println("getting geometry");
			
			//get the geometry
			VectorSet geometry = ElementProperties.getBondGeometry(atom);
			
			//now, coordinates of vectors in geometry must be transformed,
			//so that the first bonding vector coincides with the bond, and the normal of
			//the geometry is the same as the normal calculated for the parent structure
			//also, the central atom is at the position of atom
			
			geometry.transformVectorSet(atomCoordinates, normal, bond);
			
			//walk through the deficient atoms and create them using the transformed vectors
			//in geometry
			if (ProteinProperties.isAminoAcid(atom.compound)){
				//assign standard hydrogen names right here.
//				System.out.println("residue = " + atom.compound + atom.residue_id + ", atom = " + atom.name);
				try{
					HashMap m = ProteinProperties.getHNaming(atom.compound);
					String[] names = (String[])m.get(atom.name);

					for (int n = 0; n < (int)dif; n++){
						int index = n + factor;
						try{
							double[] dir = getPointOnVector(atomCoordinates, geometry.getVector(index), ElementProperties.getBondLength((short)1, atom.element));
							Atom aa = new Atom();
							aa.element = 1;
							aa.coordinate = dir;
							aa.structure = atom.structure;
							addedAtoms.put(aa, atom);
							if (names != null && n < names.length) aa.name = names[n];
							
						}
						catch (Exception ex){
//							System.err.println("residue = " + atom.compound + atom.residue_id + ", atom = " + atom.name);
//							System.out.println("names = " + names + " at " + atom.compound + ", atom name = " + atom.name + ", dif = " + dif);
							ex.printStackTrace();
							
							//use the default assignment
							
							
							
							continue;
						}
					}
				}
				catch (Exception eee){
					eee.printStackTrace();
				}
			}
			else{
				//leave atoms without names and assign them later in a general fashion
///				System.out.println("Parent atom: " + atom.name + ": assigning general names");
				for (int n = 0; n < (int)dif; n++){
					int index = n + factor;
					try{
						double[] dir = getPointOnVector(atom.coordinate, geometry.getVector(index), 1.21);
						Atom aa = new Atom();
						aa.element = 1;
						aa.coordinate = dir;
						aa.structure = atom.structure;
						addedAtoms.put(aa, atom);
					}
					catch (Exception ex){
						ex.printStackTrace();
						continue;
					}
				}
			}
		}
		else if (total > (float)valency){
			
			
			float dif = total - valency;
			if (dif < 1.0f) return;
			
//			System.out.println("dif = " + dif);
			
			//extra hydrogens are present
			//delete the extra hydrogens but take care of the geometry as well
			//the simplest way to do this is to delete all hydrogens at this atom and then add them from
			//scratch
//			System.out.println("too many hydrogens at " + atom.name);
//			System.out.println("it's h = " + atom.hybridization);
			
			
			
			//walk through all atoms around atom and delete all hydrogens
			Vector doomed = new Vector();
			Vector atoms = map.getAtoms(atom);
			//check for case when all atoms are hydrogens - then just leave one in place
			boolean allH = true;			
			for (int i = 0; i < atoms.size(); i++){
				Atom a = (Atom)atoms.get(i);
//				System.out.println("a = " + a.name);
				if (a.element == 1){
					doomed.add(a);
				}
				else{
					allH = false;
				}
			}
			
			if (allH){
				doomed.remove(atoms.get(0));//leave one atom out
			}
			
			if (doomed.size() == 0) return;//nothing to delete and subsequently check
			
			for (int j = 0; j < doomed.size(); j++){
//				System.out.println("deleted atom " + ((Atom)doomed.get(j)).name);
				viewer.deleteHydrogen((Atom)doomed.get(j), false);
			}
			
//			System.out.println("called itself");
			checkHydrogens(atom, addedAtoms, viewer, pH);
//			viewer.updateAppearance();
		}

	}
	
	/**
	 * This method is a variation of the standard ring detection method, except it can
	 * also list the Atoms that are part of the ring in the supplied Vector argument
	 * @param root
	 * @param size
	 * @param members
	 * @return
	 */
	public static boolean isInRing(Atom root, int size, Vector members){
		
		StructureMap map = root.structure.getStructureMap();
		Vector visited = new Vector();
		visited.add(root);
		Vector atoms = map.getAtoms(root);

		for (int i = 0; i < atoms.size(); i++){
			Atom a = (Atom)atoms.get(i);
			visited.clear();
			visited.add(a);
			
			//get its neighbors and start the recursion
			Vector nn = map.getAtoms(a);
			for (int j = 0; j < nn.size(); j++){
				Atom aa = (Atom)nn.get(j);
				if (aa == root) continue;
				visited.add(aa);
				
				boolean res = getNeighbors(aa, root, visited, size, 2);
				if (res){
					members = visited;
					return true;
				}
//				System.out.println("result = " + res + " + from search " + j + ", visited = " + visited.size());
			}
//			break;
		}
		
		return false;
	}

	
	/**
	 * This method determines whether a given Atom is in a ring system of the specified size
	 * @param atom
	 * @return
	 */
	public static boolean inRing(Atom root, int size){
		
		StructureMap map = root.structure.getStructureMap();
		Vector visited = new Vector();
		visited.add(root);
		Vector atoms = map.getAtoms(root);

		for (int i = 0; i < atoms.size(); i++){
			Atom a = (Atom)atoms.get(i);
			visited.clear();
			visited.add(a);
			
			//get its neighbors and start the recursion
			Vector nn = map.getAtoms(a);
			for (int j = 0; j < nn.size(); j++){
				Atom aa = (Atom)nn.get(j);
				if (aa == root) continue;
				visited.add(aa);
				
				boolean res = getNeighbors(aa, root, visited, size, 2);
				if (res){
					return true;
				}
//				System.out.println("result = " + res + " + from search " + j + ", visited = " + visited.size());
			}
//			break;
		}
		
		return false;
	}
	

	
	private static boolean getNeighbors(Atom atom, Atom root, Vector visited, int size, int position){
		
		if (position == size){
//			System.out.println("stopped traverse at " + atom.name);
			return false;
		}
		
		Vector aa = atom.structure.getStructureMap().getAtoms(atom);
//		System.err.println("Called method on atom = " + atom.name + ", aa = " + aa);

		if (aa == null){
			return false;
		}
		
		position++;
		for (int i = 0; i < aa.size(); i++){
			Atom a = (Atom)aa.get(i);
			if (a == root){
				if (position < size)return false;
				return true;
			}
			if (!visited.contains(a)){
				visited.add(a);
				boolean o = getNeighbors(a, root, visited, size, position);
				if (o){
					return true;
				}
			}
		}
		
		return false;
	}
	
	
	public static Vector getRotatedAtoms(StructureMap map, Bond bond, Atom atom){
		
		Atom next = null;//atom bonded to root
		
		//get next atom first
		if (bond.getAtom(0) == atom) next = bond.getAtom(1);
		if (bond.getAtom(1) == atom) next = bond.getAtom(0);
		
		Vector visited = new Vector();
		visited.add(atom);
		Vector atoms = map.getAtoms(atom);
		//get only the atoms on the correct side of the bond
		for (int i = 0; i < atoms.size(); i++){
			Atom a = (Atom)atoms.get(i);
			if (a == next)continue;
			if (visited.contains(a)) continue;
			visited.add(a);
			//this atom is on the correct side of the bond
			//get its neighbors and start the recursion
			Vector nn = map.getAtoms(a);
			for (int j = 0; j < nn.size(); j++){
				Atom aa = (Atom)nn.get(j);
				if (aa == atom) continue;
				if (visited.contains(aa)) continue;
				visited.add(aa);
				
				Object result = getNeighbors(aa, next, visited);
//				System.out.println("result = " + result);
				if (result == null){
					return null;
				}
			}
		}
		
		visited.remove(next);
		return visited;
	}
	
	public static Vector getNeighbors(Atom atom, Atom next, Vector visited){
		
//		System.out.println("Called method on atom = " + atom.name + ", next = " + next.name);
		Vector aa = atom.structure.getStructureMap().getAtoms(atom);
		boolean newAtoms = false;
		for (int i = 0; i < aa.size(); i++){
			Atom a = (Atom)aa.get(i);
//			System.out.println("a = " + a.name);
			if (a == next){
//				System.out.println("cycle");
				return null;
			}
			if (!visited.contains(a)){
				visited.add(a);
				Object o = getNeighbors(a, next, visited);
				if (o == null){
					return null;
				}
			}
		}
		
		return visited;
	}
	
	/**
	 * This method checks whether the structure is a protein or DNA and assigns correct bond orders in case they have not been
	 * assigned.
	 * @param structure
	 */
	public static void updateBondOrders(Structure structure){
		
		StructureMap structureMap = structure.getStructureMap();
		HashMap aaMap = ProteinProperties.getBondMap();
		HashMap naMap = DNAProperties.getBondMap();
		for (int i = 0; i < structureMap.getResidueCount(); i++){
			Residue r = structureMap.getResidue(i);
			HashMap orders = null;
			if (ProteinProperties.isAminoAcid(r.getCompoundCode())){
				//walk over the atoms and detect bond orders
				orders = (HashMap)aaMap.get(r.getCompoundCode());
			}
			else if (DNAProperties.isNucleicAcid(r.getCompoundCode())){
				orders = (HashMap)naMap.get(r.getCompoundCode());
			}
			else{
				continue;
			}
			
			for (int j = 0; j < r.getAtomCount(); j++){
				Atom a = r.getAtom(j);
				Vector atoms = structureMap.getAtoms(a);
				Vector bonds = structureMap.getBonds(a);
				if (atoms == null || bonds == null) continue;
				for (int k = 0; k < atoms.size(); k++){
					Atom aa = (Atom)atoms.get(k);
					String label = a.name + ":" + aa.name;
					if (orders.containsKey(label)){
						//get the bond that connects these two atoms
						for (int m = 0; m < bonds.size(); m++){
							Bond b = (Bond)bonds.get(m);
							if (b.getAtom(0) == a){
								if (b.getAtom(1) == aa){
									b.setOrder(((Float)orders.get(label)).floatValue());
									break;
								}
							}
							else if (b.getAtom(1) == a){
								if (b.getAtom(0) == aa){
									b.setOrder(((Float)orders.get(label)).floatValue());
									break;
								}
							}
						}
						
					}//if label is in orders
				}
			}//end of loop counting atoms
		}
		
	}
	
	public static double getRMSD(Structure structure1, Structure structure2){
		
		
		return 0;
	}
	
	/**
	 * This method performs RMSD minimization structure superposition. Two vectors of equal sizes contain Atoms
	 * that need to be compared and used to search orientation space.
	 * @param structure1
	 * @param atoms1
	 * @param structure2
	 * @param atoms2
	 */
	public static void superimpose(Structure structure1, Vector atoms1, Structure structure2, Vector atoms2){
		
		//first, determine centers of weight for both structures
		//then generate eight possible overlap orientations from the initial coordinates and choose the one
		//with the smallest rmsd.
		//using the minimum-rmsd overlap, change the coordinates by incremental rotation/translation
		//in the direction of smaller rmsd until the change is less than a given value or a maximum iteration
		//limit has been reached
		
		int max = 1000;
		double gradient = 0.1;
		
		//first, shift the second structure to the center of mass of the first
		double sumx = 0.0;
		double sumy = 0.0;
		double sumz = 0.0;
		for (int i = 0; i < atoms1.size(); i++){
			Atom a = (Atom)atoms1.get(i);
			sumx += a.coordinate[0];
			sumy += a.coordinate[1];
			sumz += a.coordinate[2];
		}
		
		double[] center1 = new double[]{ sumx/atoms1.size(), sumy/atoms1.size(), sumz/atoms1.size() };
		
		//same for the second structure
		sumx = 0.0;
		sumy = 0.0;
		sumz = 0.0;
		for (int i = 0; i < atoms2.size(); i++){
			Atom a = (Atom)atoms2.get(i);
			sumx += a.coordinate[0];
			sumy += a.coordinate[1];
			sumz += a.coordinate[2];
		}
		
		double[] center2 = new double[]{ sumx/atoms2.size(), sumy/atoms2.size(), sumz/atoms2.size() };
		
		//vector of displacement
		double[] vector = new double[]{ center1[0] - center2[0], center1[1] - center2[1], center1[2] - center2[2] };
		
		//transform coordinates of the second structure to get to the center of mass of the first one
		double[] matrix = StructureMath.getLinearTranslationMatrix(vector);
		
		for (int i = 0; i < structure2.getStructureMap().getAtomCount(); i++){
			StructureMath.transformCoordinates((structure2.getStructureMap().getAtom(i)).coordinate, matrix);
		}
		
	
		//from here, generate more orientations by flipping directional axes

		//get the translation matrices to and from the origin
		double[] toOrigin = StructureMath.getToOriginMatrix(center2);
		double[] fromOrigin = StructureMath.getFromOriginMatrix(center2);
		
		double minRMSD = 100000;
		double[] minMatrix = null;
		int minIndex = -1;
		double[] m1 = StructureMath.getLinearRotationMatrix(90, new double[]{ 1, 0, 0});		
		for (int i = 1; i <= 4; i++){
			for (int j = 0; j < atoms2.size(); j++){
				StructureMath.transformCoordinates(((Atom)atoms2.get(j)).coordinate, toOrigin);
				StructureMath.transformCoordinates(((Atom)atoms2.get(j)).coordinate, m1);
				StructureMath.transformCoordinates(((Atom)atoms2.get(j)).coordinate, fromOrigin);
			}
			
			//check the rmsd
			double rmsd = StructureMath.rmsd(atoms1, atoms2);
			if (rmsd < minRMSD){
				minRMSD = rmsd;
				minIndex = i;
			}
		}
		
		minMatrix = StructureMath.getLinearRotationMatrix(90*minIndex, new double[]{ 1, 0, 0});
		
		//check further. for now, minMatrix has the smallest rmsd.
		//if an even smaller value is found, the new minMatrix is selected, and it replaces the current one
		boolean minFound = false;
		m1 = StructureMath.getLinearRotationMatrix(90, new double[]{ 0, 1, 0});		
		for (int i = 1; i <= 4; i++){
			for (int j = 0; j < atoms2.size(); j++){
				StructureMath.transformCoordinates(((Atom)atoms2.get(j)).coordinate, toOrigin);
				StructureMath.transformCoordinates(((Atom)atoms2.get(j)).coordinate, m1);
				StructureMath.transformCoordinates(((Atom)atoms2.get(j)).coordinate, fromOrigin);
			}
			
			//check the rmsd
			double rmsd = StructureMath.rmsd(atoms1, atoms2);
			if (rmsd < minRMSD){
				minFound = true;
				minRMSD = rmsd;
				minIndex = i;
			}
		}
		
		if (minFound){
			minMatrix = StructureMath.getLinearRotationMatrix(90*minIndex, new double[]{ 0, 1, 0});
		}
		
		
		//same for the third axis
		minFound = false;
		m1 = StructureMath.getLinearRotationMatrix(90, new double[]{ 0, 0, 1});
		for (int i = 1; i <= 4; i++){
			for (int j = 0; j < atoms2.size(); j++){
				StructureMath.transformCoordinates(((Atom)atoms2.get(j)).coordinate, toOrigin);
				StructureMath.transformCoordinates(((Atom)atoms2.get(j)).coordinate, m1);
				StructureMath.transformCoordinates(((Atom)atoms2.get(j)).coordinate, fromOrigin);
			}
			
			//check the rmsd
			double rmsd = StructureMath.rmsd(atoms1, atoms2);
			if (rmsd < minRMSD){
				minFound = true;
				minRMSD = rmsd;
				minIndex = i;
			}
		}
		
		if (minFound){
			minMatrix = StructureMath.getLinearRotationMatrix(90*minIndex, new double[]{ 0, 0, 1});
		}
		
		System.out.println("min rmsd = " + minRMSD);
		
		//use the min matrix to render the best initial position
		for (int i = 0; i < structure2.getStructureMap().getAtomCount(); i++){
			Atom a = structure2.getStructureMap().getAtom(i);
			StructureMath.transformCoordinates(a.coordinate, toOrigin);
			StructureMath.transformCoordinates(a.coordinate, minMatrix);
			StructureMath.transformCoordinates(a.coordinate, fromOrigin);
		}
		
		
		//now let's start iterations to search for the minimum
		
		
		
		
		
		
		
		
		
		
		
	}
	
	public static boolean isRotatableBond(Bond bond){
		
		//this method checks whether each of the atoms of this bond is in a cycle
		//and whether they share the same cycle. also, bond order is checked
		
		if (bond.order > 1) return false;//this is not a single bond
		
		Atom atom = bond.getAtom(0);
		Atom next = bond.getAtom(1);
		
		StructureMap map = atom.structure.getStructureMap();
		
		Vector visited = new Vector();
		visited.add(atom);
		Vector atoms = map.getAtoms(atom);
		
		for (int i = 0; i < atoms.size(); i++){
			Atom a = (Atom)atoms.get(i);
			if (a == next)continue;
			if (visited.contains(a)) continue;
			visited.add(a);
			
			//this atom is on the correct side of the bond
			//get its neighbors and start the recursion
			Vector nn = map.getAtoms(a);
			for (int j = 0; j < nn.size(); j++){
				Atom aa = (Atom)nn.get(j);
				if (aa == atom) continue;
				if (visited.contains(aa)) continue;
				visited.add(aa);
				
				Object result = StructureMath.getNeighbors(aa, next, visited);

				if (result == null){
					return false;
				}
			}
		}

		
		
		return true;
		
	}

	
	
	public static double rmsd(Vector atoms1, Vector atoms2){
		
		if (atoms1.size() != atoms2.size()) return -1.0;
		
		double sum = 0.0;
		for (int i = 0; i < atoms1.size(); i++){
			Atom a = (Atom)atoms1.get(i);
			Atom b = (Atom)atoms2.get(i);
			
			double dist = Algebra.distance(a.coordinate, b.coordinate);
			sum += (dist * dist);
		}
		
		return Math.sqrt(sum);
	}
	
	
	
	public static int countHydrogenBonds(Structure structure1, Structure structure2){
		
		//get the Atom vectors
		Vector ligand = structure1.getStructureMap().getAtoms();
		Vector protein  = structure2.getStructureMap().getAtoms();
		double maximum = 3;
		
		int counter = 0;
		
		//first, find hydrogens with heteroatoms on both sets
		for (int i = 0; i < ligand.size(); i++){
			Atom a = (Atom)ligand.get(i);
			if (a.element == 1){
				try{
					Atom next = (Atom)(a.structure.getStructureMap().getAtoms(a)).get(0);
					if (next == null) continue;
					if (next.element == 8 || next.element == 7 || next.element == 16){
						//scan through the atoms of protein set to find heteroatoms
						for (int j = 0; j < protein.size(); j++){
							Atom aa = (Atom)protein.get(j);
							if (aa == a || aa == next) continue;
							if (ligand.contains(aa)) continue;
							if (aa.element == 8 || aa.element == 7 || aa.element == 16){
								if (!aa.visible) continue;
								if (Algebra.distance(a.coordinate, aa.coordinate) <= maximum){
									boolean pass = false;
									
									//a possible H bond
									//check its geometry to make sure the vectors are not at 
									//a sharp angle
									double angle1 = Algebra.angle(next.coordinate, a.coordinate, aa.coordinate);
									if (angle1 < 130) continue;
									
									double angle2 = 0;
									//angle2 is between H, acceptor heteroatom, and the base of the acceptor heteroatom
									Vector atoms = aa.structure.getStructureMap().getAtoms(aa);
									if (atoms == null) continue;
									if (atoms.size() == 0) continue;
									if (atoms.size() == 1){
										//only one bond (possibly double)
										angle2 = Algebra.angle(a.coordinate, aa.coordinate, ((Atom)atoms.get(0)).coordinate);
										if (angle2 > 110){
											if (angle1 < 120 || angle2 < 120){
												if (Algebra.distance(a.coordinate, aa.coordinate) > 2.5){
													continue;
												}
											}
											pass = true;
										}
									}
									else{
										double[] h = new double[]{ aa.coordinate[0] - a.coordinate[0], aa.coordinate[1] - a.coordinate[1], aa.coordinate[2] - a.coordinate[2] };
										if (atoms.size() == 2){
											Atom a1 = (Atom)atoms.get(0);
											Atom a2 = (Atom)atoms.get(1);
											
											double[] b = new double[]{ a1.coordinate[0] - aa.coordinate[0],a1.coordinate[1] - aa.coordinate[1],a1.coordinate[2] - aa.coordinate[2] }; 
											double[] aux = new double[]{ a1.coordinate[0] - a2.coordinate[0], a1.coordinate[1] - a2.coordinate[1], a1.coordinate[2] - a2.coordinate[2] };
											Algebra.normalizeVector(aux);
											double[] normal = Algebra.crossProduct(b, aux);
											Algebra.normalizeVector(normal);
											
											double[] pointer = Algebra.crossProduct(aux, normal);
											//check the direction: look at the angle with b, should be greater than 90 degrees
											double angle = Algebra.getAngleBetweenVectors(pointer, b);
											if (angle < 100){
												Algebra.reverseVector(pointer);
											}
											
											//get the vector from hydrogen to this heteroatom and get the angle
											angle2 = Algebra.getAngleBetweenVectors(h, pointer);
											
											if (angle2 > 130) pass = true;
										}
									}
									if (pass){
										counter++;
									}
									
								}
							}
						}
					}
				}
				catch (Exception e){
					continue;
				}
			}
		}
		
		for (int i = 0; i < protein.size(); i++){
			Atom a = (Atom)protein.get(i);
			if (a.element == 1){
				Atom next = (Atom)(a.structure.getStructureMap().getAtoms(a)).get(0);
				if (next == null) continue;
				if (next.element == 8 || next.element == 7 || next.element == 16){
					//scan through the atoms of ligand set to find heteroatoms
					for (int j = 0; j < ligand.size(); j++){
						Atom aa = (Atom)ligand.get(j);
						if (aa.element == 8 || aa.element == 7 || aa.element == 16){
							if (aa.compound.equals("HOH")) continue;
							if (Algebra.distance(a.coordinate, aa.coordinate) <= maximum){
								//a possible H bond
								//check its geometry to make sure the vectors are not at 
								//a sharp angle
								double angle1 = Algebra.angle(next.coordinate, a.coordinate, aa.coordinate);
								
								if (angle1 < 130) continue;
								
								boolean pass = false;
								
								double angle2 = 0;
								//angle2 is between H, acceptor heteroatom, and the base of the acceptor heteroatom
								Vector atoms = aa.structure.getStructureMap().getAtoms(aa);
								if (atoms == null) continue;
								if (atoms.size() == 0) continue;
								if (atoms.size() == 1){
									//only one bond (possibly double)
									angle2 = Algebra.angle(a.coordinate, aa.coordinate, ((Atom)atoms.get(0)).coordinate);
//									System.out.println(a.name + "-" + aa.name + "-" + ((Atom)atoms.get(0)).name + ": angle2 = " + angle2);
									if (angle2 > 110){
										if (angle1 < 120 || angle2 < 120){
											if (Algebra.distance(a.coordinate, aa.coordinate) > 2.5){
//												System.out.println("Sharp angles: rejected");
												continue;
											}
										}
										pass = true;
									}
								}
								else{
									double[] h = new double[]{ aa.coordinate[0] - a.coordinate[0], aa.coordinate[1] - a.coordinate[1], aa.coordinate[2] - a.coordinate[2] };
									if (atoms.size() == 2){
//										System.out.println("2 bonds for " + aa.name + " in " + aa.compound);
										Atom a1 = (Atom)atoms.get(0);
										Atom a2 = (Atom)atoms.get(1);
										
										double[] b = new double[]{ a1.coordinate[0] - aa.coordinate[0],a1.coordinate[1] - aa.coordinate[1],a1.coordinate[2] - aa.coordinate[2] }; 
										double[] aux = new double[]{ a1.coordinate[0] - a2.coordinate[0], a1.coordinate[1] - a2.coordinate[1], a1.coordinate[2] - a2.coordinate[2] };
										Algebra.normalizeVector(aux);
										double[] normal = Algebra.crossProduct(b, aux);
										Algebra.normalizeVector(normal);
										
										double[] pointer = Algebra.crossProduct(aux, normal);
										//check the direction: look at the angle with b, should be greater than 90 degrees
										double angle = Algebra.getAngleBetweenVectors(pointer, b);
										if (angle < 100){
											Algebra.reverseVector(pointer);
										}
										
//										this.createArrow(aa.structure, aa.coordinate, StructureMath.getPointOnVector(aa.coordinate, pointer, 1), new float[]{ 1.0f, 0.0f, 0.0f });
										
										//get the vector from hydrogen to this heteroatom and get the angle
										angle2 = Algebra.getAngleBetweenVectors(h, pointer);
//										System.out.println("angle2 = " + angle2);
										
										if (angle2 > 130) pass = true;
									}
								}

								
								if (pass){
									counter++;
								}	
								
								
							}
						}
					}

			
				}
			}
		}

		
		return counter;
	}

	
}
