//Copyright (c) 2000-2003  San Diego Supercomputer Center (SDSC),
//a facility operated by the University of California, San Diego (UCSD)
//
//Users and possessors of this source code are hereby granted a
//nonexclusive, royalty-free copyright and design patent license to
//use this code in individual software.  License is not granted for
//commercial resale, in whole or in part, without prior written
//permission from SDSC.  This source is provided "AS IS" without express
//or implied warranty of any kind.
//
//For further information, please see:  http://mbt.sdsc.edu
//
package edu.sdsc.sirius.io;


import edu.sdsc.mbt.*;
import edu.sdsc.mbt.util.*;
import edu.sdsc.sirius.util.DNAProperties;
import edu.sdsc.sirius.util.ProteinProperties;

import java.util.*;
import java.io.*;
import java.net.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.zip.*;

public class PdbEntryLoader implements FileEntryLoader, java.io.Serializable {

	protected String urlString;
	private long expectedInputBytes = 1;
	
	private String currentName;
	
	private String[] chainIdList = new String[]{ "A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M", 
			"N", "O", "P", "Q", "R", "S", "T", "U", "V", "W", "X", "Y", "Z" };

	
	private final int DATA_TYPE = Entry.DATA_STRUCTURE;

	// A hashtable of vectors where
	// each hash KEY is the StructureComponent type String.
	// each hash VALUE is a Vector of StructureComponent objects.
	// NOTE: This protected variable is used only to enable each new
	// Structure object to access the newly loaded data. The Structure
	// constructor stores a copy of the object so this variable can
	// be re-used for the next load request. This is not thread-safe
	// but there doesn't seem to be any other way to enable an
	// anonymous class to be handed state (aside from parameters).
	protected Hashtable passComponents;


	//
	// StructureLoader interface methods
	//
	/**
	 * Returns name of the loader.
	 */
	public String getLoaderName(){
		return new String("Format PDB");
	}

	/**
	 * Loads the Entry by file name.
	 */
	public Entry load( String name, int number ){
	
		StructureEntry entry = new StructureEntry();
		
		Structure structure = null;//future structureObject derived from the PDB file
		try
		{	
			File file = new File( name );
			structure = loadStructure( file, number );
		}
		catch( NullPointerException e )
		{
			try
			{	
				URL url = new URL( name );
				structure = loadStructure( url, number );
				entry.setSourceUrl(url);
			}
			catch( MalformedURLException e2 )
			{
			}
		}
		if ( structure == null ){
			return null;
		}
		
		
		String tag = null;
		try{
			String temp = name;
			int index = temp.lastIndexOf("/") + 1;
			String tUrlString = temp.substring(index);
			index = tUrlString.lastIndexOf(".");
			tag = tUrlString.substring(0,index);
		}
		catch (Exception ex){
			tag = structure.getUrlString();
		}
		
		
		urlString = name;
		
		//modify the file name to get rid of the extension and use this string as the entry name
		if (currentName == null){
			String shortName = name;
			if (name.endsWith(".pdb") || name.endsWith(".PDB")){
				shortName = name.substring(0, name.length()-4);
			}
			else if (name.endsWith(".pdb.gz")){
				shortName = name.substring(0, name.length()-7);
			}
			
			if(shortName.indexOf(File.separator) > -1){
				int index = shortName.lastIndexOf(File.separator);
				String temp = shortName.substring(index+1);
				shortName = temp;
			}
			entry.setName(shortName);
		}
		else{
			tag += "$" + currentName;
			entry.setName(tag);
		}
		
		entry.setStructure(structure);
		
		
		return entry;
	}
	
	public Vector readHeaders(URL url){
		
		try
		{

			URLConnection urlConnection = url.openConnection( );
			InputStream inputStream = urlConnection.getInputStream( );
			
			BufferedInputStream bufferedInputStream = null;
			if ( inputStream != null ) urlString = url.toExternalForm( );
			if ( url.toExternalForm().endsWith( ".pdb.gz" ) )
			{
				GZIPInputStream gzipInputStream =
					new GZIPInputStream( inputStream );
				bufferedInputStream =
					new BufferedInputStream( gzipInputStream );
			}
			else
			{
				bufferedInputStream =
					new BufferedInputStream( inputStream );
			}
			return readHeaders( bufferedInputStream );
		}
		catch ( FileNotFoundException e )
		{
			e.printStackTrace();
			return null;
		}
		catch ( SecurityException e )
		{
			return null;
		}
		catch (Exception ex){
			ex.printStackTrace();
			return null;
		}

	}

	
	private Vector readHeaders(BufferedInputStream bufferedInputStream){
		
		Vector out = new Vector();
		
		if ( bufferedInputStream == null ) return null;

		// System.err.println( "PdbStructureLoader.load: expectedBytes = " + expectedBytes );

		byte buf[] = new byte[ 2048 ]; // raw read data buffer
		byte line[] = new byte[ 100 ]; // accumulated line buffer

		
		int linePos = 0;
		int bytesRead = 0;
		int lastRead = 0;
		int lines = 0;
		int readCount = 0;

		int lineNumber = 0;
		boolean model = false;
		
		try
		{
			while ( bufferedInputStream.available() > 0 )
			{
				// Re-fill the raw data buffer.
				try
				{
					lastRead = bufferedInputStream.read( buf, 0, buf.length );
				}
				catch( IOException e )
				{
					System.err.println( "PDBEntryLoader.readHeaders: " + e );
					return null;
				}
				bytesRead += lastRead;

				// Process the buffer
				for ( int bufPos=0; bufPos<lastRead; bufPos++ )
				{
					// Copy a byte from the input buffer into the line buffer
					line[linePos] = buf[bufPos];
					if ( buf[bufPos] == '\n' )
					{
						// line[0-linePos] now has a complete line.
						lines++;
						linePos = 0;
						lineNumber++;
						
						
						if (
								(line[0] == 'M') &&
								(line[1] == 'O') &&
								(line[2] == 'D') &&
								(line[3] == 'E') &&
								(line[4] == 'L') 
						){

							//this is the first line of a molecule record: get the name
							currentName = new String(line);
							currentName = currentName.substring(0, currentName.indexOf("\n")).trim();
							currentName.trim();
							out.add(currentName);
							model = true;
							continue;
						}
						
						
						if (
								(line[0] == 'A') &&
								(line[1] == 'T') &&
								(line[2] == 'O') &&
								(line[3] == 'M') 
						){
							if (!model){
								//atoms started without a model flag
								return null;
							}
						}
						
						if (
								(line[0] == 'H') &&
								(line[1] == 'E') &&
								(line[2] == 'T') &&
								(line[3] == 'A') &&
								(line[4] == 'T') &&
								(line[5] == 'M') 
						){
							if (!model){
								//atoms started without a model flag
								return null;
							}
						}
						
					}
					else
					{
						// Continue building the line.
						linePos++;
						if ( linePos >= line.length )
						{
							System.err.println( "dpassement de capacit du tampon de ligne !" );
							return null;
						}
					}
				}
			}
		}
		catch( Exception e )
		{
			e.printStackTrace();
			return null;
		}


		
		return out;
	}
	
	
	/**
	 * Loads the Entry from a File object.
	 */
	public Entry load(File file, int number){
		
		StructureEntry entry = new StructureEntry();
		
		String name = file.toString();
//		urlString = name;
		
		Structure structure = null;
		try {	
			structure = loadStructure( file, number );
			entry.setSourceUrl(file.toURL());
		}
		catch (NumberFormatException ex){
			throw ex;
		}
		catch( Exception e ){
			e.printStackTrace();
		}
		
		if ( structure == null ){
			return null;
		}

		String tag = null;
		try{
			String temp = file.toURL().toExternalForm();
			int index = temp.lastIndexOf("/") + 1;
			String tUrlString = temp.substring(index);
			index = tUrlString.lastIndexOf(".");
			tag = tUrlString.substring(0,index);
		}
		catch (Exception ex){
			tag = structure.getUrlString();
		}
		
		
		//modify the file name to get rid of the extension and use this string as the entry name
		if (currentName == null){
			String shortName = name;
			if (name.endsWith(".pdb") || name.endsWith(".PDB")){
				shortName = name.substring(0, name.length()-4);
			}
			else if (name.endsWith(".pdb.gz")){
				shortName = name.substring(0, name.length()-7);
			}
	
			if(shortName.indexOf(File.separator) > -1){
				int n = shortName.lastIndexOf(File.separator);
				String temp2 = shortName.substring(n+1);
				shortName = temp2;
			}
			entry.setName(shortName);

		}
		else{
			tag += "$" + currentName;
			entry.setName(tag);
		}
		
		entry.setStructure(structure);
		
		return entry;
	
	
	}
	
	/**
	 * Loads the Entry from a URL object.
	 */
	public Entry load(URL url, int number){
	
		StructureEntry entry = new StructureEntry();
		
		String name = url.toString();
		urlString = name;
		
		Structure structure = loadStructure( url, number );
		
		if ( structure == null ){
			return null;
		}
		
		if (structure.getStructureMap().getResidueCount() == 0) return null;
		
		String tag = null;
		try{
			String temp = url.toExternalForm();
			int index = temp.lastIndexOf("/") + 1;
			String tUrlString = temp.substring(index);
			index = tUrlString.lastIndexOf(".");
			tag = tUrlString.substring(0,index);
		}
		catch (Exception ex){
			tag = structure.getUrlString();
		}

		entry.setSourceUrl(url);
		
		//modify the file name to get rid of the extension and use this string as the entry name
		if (currentName == null){
			String shortName = name;
			if (name.endsWith(".pdb") || name.endsWith(".PDB")){
				shortName = name.substring(0, name.length()-4);
			}
			else if (name.endsWith(".pdb.gz")){
				shortName = name.substring(0, name.length()-7);
			}
	
			if(shortName.indexOf(File.separator) > -1){
				int index = shortName.lastIndexOf(File.separator);
				String temp = shortName.substring(index+1);
				shortName = temp;
			}
			else if(shortName.indexOf("/") > -1){
				int index = shortName.lastIndexOf("/");
				String temp = shortName.substring(index+1);
				shortName = temp;
			}
			entry.setName(shortName);
		}
		else{
			tag += "$" + currentName;
			entry.setName(tag);
		}

		entry.setStructure(structure);
		
		return entry;
	}

	
	/**
	 * Returns true if the loader is capable of loading the structure,
	 * or false otherwise. This enables higher-level code to be able
	 * to build a context sensative menu of only the loaders that can
	 * load a given structure name.
	 */
	public boolean canLoad( String name )
	{
		// System.err.println( "PdbStructureLoader.canLoad(String)" );
		if ( ! name.endsWith( ".pdb" ) && ! name.endsWith(".PDB") && ! name.endsWith( ".pdb.gz" ) )
			return false;

		return true;
	}


	/**
	 * Returns true if the Entry can be read from the given File object.
	 */
	public boolean canLoad( File file )
	{
		// System.err.println( "PdbStructureLoader.canLoad(File)" );
		if ( ! file.exists() ) return false;

		try
		{
			// String path = file.getCanonicalPath( );
			String path = file.toURL().toExternalForm();
			return canLoad( path );
		}
		catch( IOException e )
		{
			return false;
		}
	}
	
	public Vector readHeaders(File file){
		
		long expectedInputBytes = file.length( );
		if ( expectedInputBytes <= 0 ) return null;

		try
		{
			FileInputStream fileInputStream = new FileInputStream( file );
			if ( fileInputStream == null ) return null;
			
			//filename was being reduced to just the name itself here and in SequenceViewerPanel
//			urlString = file.toURL().toExternalForm();
			
			String tempUrlString = null;
			tempUrlString = file.toURL().toExternalForm();
			
			//process the name to get rid of the full path and keep
			//only the file name without the extension for internal reference purposes
			int index = tempUrlString.lastIndexOf("/") + 1;
			String tUrlString = tempUrlString.substring(index);
		
			index = tUrlString.lastIndexOf(".");
			
			
			BufferedInputStream bufferedInputStream = new BufferedInputStream( fileInputStream );

			return readHeaders( bufferedInputStream, file.length() );
		}
		catch ( FileNotFoundException e )
		{
			return null;
		}
		catch ( SecurityException e )
		{
			return null;
		}
		catch (Exception ex){
			return null;
		}
	}

	private Vector readHeaders(BufferedInputStream bufferedInputStream, long expected){
		
		Vector out = new Vector();
		
		if ( bufferedInputStream == null ) return null;

		// System.err.println( "PdbStructureLoader.load: expectedBytes = " + expectedBytes );

		byte buf[] = new byte[ 2048 ]; // raw read data buffer
		byte line[] = new byte[ 100 ]; // accumulated line buffer

		
		int linePos = 0;
		int bytesRead = 0;
		int lastRead = 0;
		int lines = 0;
		int readCount = 0;

		int lineNumber = 0;
		boolean model = false;
		
		try
		{
			while ( bufferedInputStream.available() > 0 )
			{
				// Re-fill the raw data buffer.
				try
				{
					lastRead = bufferedInputStream.read( buf, 0, buf.length );
				}
				catch( IOException e )
				{
					System.err.println( "PDBEntryLoader.readHeaders: " + e );
					return null;
				}
				bytesRead += lastRead;

				// Process the buffer
				for ( int bufPos=0; bufPos<lastRead; bufPos++ )
				{
					// Copy a byte from the input buffer into the line buffer
					line[linePos] = buf[bufPos];
					if ( buf[bufPos] == '\n' )
					{
						// line[0-linePos] now has a complete line.
						lines++;
						linePos = 0;
						lineNumber++;
						
						
						if (
								(line[0] == 'M') &&
								(line[1] == 'O') &&
								(line[2] == 'D') &&
								(line[3] == 'E') &&
								(line[4] == 'L') 
						){

							//this is the first line of a molecule record: get the name
							currentName = new String(line);
							currentName = currentName.substring(0, currentName.indexOf("\n")).trim();
							currentName.trim();
							out.add(currentName);
							model = true;
							continue;
						}
						
						
						if (
								(line[0] == 'A') &&
								(line[1] == 'T') &&
								(line[2] == 'O') &&
								(line[3] == 'M') 
						){
							if (!model){
								//atoms started without a model flag
								return null;
							}
						}
						
						if (
								(line[0] == 'H') &&
								(line[1] == 'E') &&
								(line[2] == 'T') &&
								(line[3] == 'A') &&
								(line[4] == 'T') &&
								(line[5] == 'M') 
						){
							if (!model){
								//atoms started without a model flag
								return null;
							}
						}
						
					}
					else
					{
						// Continue building the line.
						linePos++;
						if ( linePos >= line.length )
						{
							System.err.println( "dpassement de capacit du tampon de ligne !" );
							return null;
						}
					}
				}
			}
		}
		catch( Exception e )
		{
			e.printStackTrace();
			return null;
		}


		
		return out;
	}



	/**
	 * Returns a reference to a Structure read from the given URL object.
	 */
	public Structure loadStructure( URL url, int number )
	{
		// System.err.println( "PdbStructureLoader.load(URL)" );

		if ( ! canLoad( url ) ) return null;

		try
		{
			URLConnection urlConnection = url.openConnection( );
			expectedInputBytes = urlConnection.getContentLength( );
//			System.out.println("bytes = " + expectedInputBytes);
//			if ( expectedInputBytes <= 0 ) return null;
			InputStream inputStream = urlConnection.getInputStream( );
			
			BufferedInputStream bufferedInputStream = null;
			if ( inputStream != null ) urlString = url.toExternalForm( );
			if ( url.toExternalForm().endsWith( ".pdb.gz" ) )
			{
				GZIPInputStream gzipInputStream =
					new GZIPInputStream( inputStream );
				bufferedInputStream =
					new BufferedInputStream( gzipInputStream );
			}
			else
			{
				bufferedInputStream =
					new BufferedInputStream( inputStream );
			}
			return loadStructure( bufferedInputStream, number );
		}
		catch( IOException e )
		{
			return null;
		}
	}
	
	public Structure loadStructure( File file, int number )
	{
		// System.err.println( "PdbStructureLoader.load(File)" );
		
//		System.out.println("loadStructure with id = " + identificator);

		if ( ! canLoad( file ) ) return null;
		if ( ! file.exists() ) return null;
		if ( ! file.canRead() ) return null;

		expectedInputBytes = file.length( );
		if ( expectedInputBytes <= 0 ) return null;

		try
		{
			FileInputStream fileInputStream = new FileInputStream( file );
			if ( fileInputStream == null ) return null;
			
			//filename was being reduced to just the name itself here and in SequenceViewerPanel
//			urlString = file.toURL().toExternalForm();
			
			String tempUrlString = null;
			tempUrlString = file.toURL().toExternalForm();
			
			//process the name to get rid of the full path and keep
			//only the file name without the extension for internal reference purposes
			int index = tempUrlString.lastIndexOf("/") + 1;
			String tUrlString = tempUrlString.substring(index);
		
			index = tUrlString.lastIndexOf(".");
			urlString = tUrlString.substring(0,index);
			
			BufferedInputStream bufferedInputStream = null;
			if ( file.toString().endsWith( ".gz" ) )
			{
				GZIPInputStream gzipInputStream =
					new GZIPInputStream( fileInputStream );
				bufferedInputStream =
					new BufferedInputStream( gzipInputStream );

				// JLM DEBUG: crude hack for progress, because with
				// a gzip stream we can't tell how much data there will be!
				// A 4:1 compression is typical for PDB files.
				expectedInputBytes *= 4;
			}
			else
			{
				bufferedInputStream =
					new BufferedInputStream( fileInputStream );
			}
			return loadStructure( bufferedInputStream, number );
		}
		catch (NumberFormatException e){
			throw e;
		}
		catch ( FileNotFoundException e )
		{
			return null;
		}
		catch ( MalformedURLException e )
		{
			return null;
		}
		catch ( SecurityException e )
		{
			return null;
		}
		catch ( IOException e )
		{
			return null;
		}
	}



	/**
	 * Returns true if the Entry can be read from the given URL object.
	 */
	public boolean canLoad( URL url )
	{
		// System.err.println( "PdbStructureLoader.canLoad(URL)" );
		return canLoad( url.toExternalForm( ) );
	}

	//
	// PdbStructureLoader methods
	//

	/**
	 * Returns a reference to a Structure read from the given (uncompressed)
	 * InputStream.
	 */
	public Structure loadStructure( BufferedInputStream bufferedInputStream, int number ){
		
		if ( bufferedInputStream == null ) return null;
		
//		System.out.println("Loading");

		passComponents = new Hashtable( );
		long expectedBytes = expectedInputBytes;

		byte buf[] = new byte[ 4096 ]; // raw read data buffer
		byte line[] = new byte[ 100 ]; // accumulated line buffer

		int onePercent = (int) (((float) expectedBytes / (float) buf.length) / 100.f);
		if ( onePercent <= 0 ) onePercent = 1;
		float percentDone = 0.0f;

		if (Status.enabled) Status.progress( percentDone, "Chargement..." );
		
		int[][] conect = null;
		int hetatmCount = 0;
		
		Vector hetatms = new Vector();//the heteroatom numbers are stored in order they are in the file
		
		HashMap numberOrder = new HashMap();//the other way around: atom number -> position
		
		int linePos = 0;
		int bytesRead = 0;
		int lastRead = 0;
		int lines = 0;
		int readCount = 0;
		
		HashMap extraBonds = new HashMap();//root Atom -> array of atoms that are in bonds and non-regular bonds in the protein
		//stores Atom references
		
		boolean protein = false;
		boolean dna = false;
		
		//aminoacid parsing variables
		Atom lastC = null;//C-terminal carbon of the last residue (to connect to the next N)
		Atom currentC = null;
		
		Atom lastO = null;//for DNA
		Atom currentO = null;
		
		int lastResidueId = -1;//to track change of residues
		String lastResidueName = null;
		String lastChainId = null;
		String currentChainId = null;
		boolean chainChange = false;
		HashMap residueAtoms = new HashMap();//stores name->Atom pairs for the current residue
		int dif = 1;
		
		Vector ter = new Vector();//positions of TER records to be used in case there is no explicit chain labels
		
		//frequently, there is no explicit chain id, but there are TER records that delimit separate chains.
		//here, if there are no chain ids in the file, automatically assign id A, and increment to B (and so on) as TER records
		//are seen.
		//if no TER records are in the file, revert all atoms to the default "_"
		int chainIdCounter = 0;
		
//		System.out.println("expectedBytes = " + expectedBytes);
		
		int modelCounter = 0;
		boolean load = false;
		
		if (number == -1) load = true;//no models - just straight load
		
		currentName = null;
		
		boolean waterStarted = false;
		
		boolean stop = false;
		
		int newline = -1;//index of newline in the current line
		
		try
		{
			while ( bufferedInputStream.available() > 0 )
			{
				
				if (stop){
					bufferedInputStream.close();
					break;
				}
				
//				System.out.println("new pass");
				// Re-fill the raw data buffer.
				try
				{
					lastRead = bufferedInputStream.read( buf, 0, buf.length );
				}
				catch( IOException e )
				{
					System.err.println( "PdbEntryLoader.load: " + e );
					return null;
				}
				bytesRead += lastRead;

				readCount++;
				if ( readCount % onePercent == 0 ){
					percentDone = (float) bytesRead / (float) expectedBytes;
					if (Status.enabled) Status.progress( percentDone, "Chargement " + urlString );
				}
	

				// Process the buffer
				for ( int bufPos=0; bufPos<lastRead; bufPos++ )
				{
					// Copy a byte from the input buffer into the line buffer
					line[linePos] = buf[bufPos];
					if ( buf[bufPos] == '\n' )
					{
						// line[0-linePos] now has a complete line.
						lines++;
	
						// Parse the line buffer
//						System.out.println(new String(line));
						
						
						if (
							(line[0] == 'M') &&
							(line[1] == 'O') &&
							(line[2] == 'D') &&
							(line[3] == 'E') &&
							(line[4] == 'L')
						){
							if (modelCounter == number){
								currentName = new String(line);
								currentName = currentName.substring(0, currentName.indexOf("\n")).trim();
								currentName.trim();
								
								//get rid of any duplicate spaces (contiguous spaces)
								Pattern pattern = Pattern.compile("\\s+");
								Matcher matcher = pattern.matcher(currentName);
								String compressed = matcher.replaceAll("_");
								currentName = compressed;
								load = true;
							}
							modelCounter++;
							linePos = 0;
							continue;
						}
						
						if (
								(line[0] == 'E') &&
								(line[1] == 'N') &&
								(line[2] == 'D') &&
								(line[3] == 'M') &&
								(line[4] == 'D') &&
								(line[5] == 'L')
							){
								if (load){
									linePos = 0;
									load = false;
								}
							}
							
						if (!load){
							linePos = 0;
							continue;
						}
						
						if (
								(line[0] == 'E') &&
								(line[1] == 'N') &&
								(line[2] == 'D') &&
								(line[3] == '\n')
							){
								//finish
								stop = true;
								break;
							}

	
						//
						// ATOM and HETATM records
						//
						boolean isAtom = false;
						boolean isHetatm = false;
						if (
							(line[0] == 'A') &&
							(line[1] == 'T') &&
							(line[2] == 'O') &&
							(line[3] == 'M')
						)
							isAtom = true;
						else if (
							(line[0] == 'H') &&
							(line[1] == 'E') &&
							(line[2] == 'T') &&
							(line[3] == 'A') &&
							(line[4] == 'T') &&
							(line[5] == 'M')
						){
							isHetatm = true;
						}
						
						if (
								(line[0] == 'T') &&
								(line[1] == 'E') &&
								(line[2] == 'R') 
							){
							
							Vector records = (Vector) passComponents.get(StructureComponentRegistry.TYPE_ATOM );
							if ( records != null ){
								ter.add(records.get(records.size()-1));//last Atom before the TER record
							}
						}
						
						if ( isAtom || isHetatm)
						{
							// PDB File Atom Record offsets as documented by
							// http://www.rcsb.org/pdb/docs/format/pdbguide2.2/part_62.html
 							// 1 -  6   RecordName
 							// 7 - 11   serial
 							// 12       -
							// 13 - 16  name
							// 17       altLoc
							// 18 - 20  resName
							// 21       -
							// 22       chainID
							// 23 - 26  resSeq
							// 27       iCode
							// 28 - 30  -
							// 31 - 38  x
							// 39 - 46  y
							// 47 - 54  z
							// 55 - 60  occupancy
							// 61 - 66  tempFactor
							// 67 - 72  -
							// 73 - 76  segID
							// 77 - 78  element
							// 79 - 80  charge
							// NOTE: In this application, we need to subtract 1 from
							// each index in order to match the 0-based array offsets.
							
							
							Atom atom = new Atom( );
							String str = null;

							atom.number =  Integer.parseInt( (new String( line, 6, 5 )).trim() );
							if (isHetatm){
//								orderNumber.put(new Integer(hetatmCount), new Integer(atom.number));
								numberOrder.put(new Integer(atom.number), new Integer(hetatmCount));
//								System.out.println("index = " + hetatmCount + ", number = " + atom.number);
								hetatms.add(atom);
								hetatmCount++;
							}
							
							atom.name = (new String( line, 12, 4 )).trim();
							
							if (linePos > 77){
								String ene = (new String( line, 77, 2 )).trim();
								atom.setElement(ene);
								if (!PeriodicTable.isElement(atom.element)){
									//look elsewhere for element name
									String en2 = (new String( line, 12, 2 )).trim();
									if (PeriodicTable.isElementSymbol(en2)){
										atom.setElement(en2);
									}
									else{
										//check if there is a leading number
										atom.setElement(en2.substring(1));
										if (!PeriodicTable.isElementSymbol(en2.substring(1))){
											atom.element = 6;
										}
									}
								}
							}
							else{
								String en2 = (new String( line, 12, 2 )).trim();
								if (PeriodicTable.isElementSymbol(en2)){
									atom.setElement(en2);
								}
								else{
									//check if there is a leading number
									atom.setElement(en2.substring(1));
									if (!PeriodicTable.isElementSymbol(en2.substring(1))){
										atom.element = 6;
									}
								}
							}
							
//							System.out.println("name = " + atom.name);
							atom.compound = (new String( line, 17, 3 )).trim();
							if (atom.compound == null) atom.compound = "UKN";
							
							//check for starting waters
							if (!waterStarted){
								if(atom.compound.equals("HOH") || atom.compound.equals("WAT")){
									waterStarted = true;
									Vector records = (Vector) passComponents.get(StructureComponentRegistry.TYPE_ATOM );
									if ( records != null ){
										//if ter record has not been found in the file, add one
										Object last = records.get(records.size()-1);
										if (!ter.contains(last)){
											ter.add(last);
										}
									}

								}
							}
							
							atom.chain_id = (new String( line, 21, 1 )).trim();
							if ( atom.chain_id == null ){
								atom.chain_id = "_";
								
								//try to get the residue number from differently placed fields
								try{
									atom.residue_id =	Integer.parseInt( (new String( line, 22, 4 )).trim() );
								}
								catch (Exception ex){
									atom.residue_id =	Integer.parseInt( (new String( line, 23, 4 )).trim() );
								}
							}
							else if ( atom.chain_id.equals( "" ) ){
								atom.chain_id = "_";
								
								//try to get the residue number from differently placed fields
								try{
									atom.residue_id =	Integer.parseInt( (new String( line, 22, 4 )).trim() );
								}
								catch (Exception ex){
									try{
										atom.residue_id =	Integer.parseInt( (new String( line, 23, 4 )).trim() );
									}
									catch (Exception e){
										atom.residue_id = 1;
									}
								}
							}
							else{
								//try to get the residue number from differently placed fields
								String input = (new String( line, 22, 5 )).trim();
								try{
									atom.residue_id =	Integer.parseInt(input);
								}
								catch (Exception ex){
									try{
										atom.residue_id =	Integer.parseInt( input );
									}
									catch (NumberFormatException e){
										//if input contains letters in addition to numbers, use only the numeric part
										try{
											Pattern pattern = Pattern.compile("[A-Za-z]");
											Matcher matcher = pattern.matcher(input);
											atom.residue_id = Integer.parseInt(matcher.replaceAll(""));
										}
										catch (NumberFormatException ee){
											atom.residue_id = 1;
										}
									}
								}
							}
							
							atom.coordinate = new double[3];
							atom.coordinate[0] =
								Double.parseDouble( (new String( line, 30, 8 )).trim() );
							atom.coordinate[1] =
								Double.parseDouble( (new String( line, 38, 8 )).trim() );
							atom.coordinate[2] =
								Double.parseDouble( (new String( line, 46, 8 )).trim() );


							Vector records = (Vector) passComponents.get(
								StructureComponentRegistry.TYPE_ATOM );
							if ( records == null )
							{
								records = new Vector( );
								passComponents.put(
									StructureComponentRegistry.TYPE_ATOM, records );
							}
							records.add( atom );
//							System.out.println("added atom");
														
							try{
//								System.out.println("compound = " + atom.compound);
								if (ProteinProperties.isAminoAcid(atom.compound)){
									if (atom.residue_id != lastResidueId){
										//there was a switch of residue or this is only the start of the chain
										if (lastResidueName != null){
											//check if this is a chain change
											//also, this may be needed for gaps in residue numbering
											//to avoid making long bonds that span the gap
											if (lastC != null){
												dif = currentC.residue_id - lastC.residue_id;
											}
											if (lastChainId != null){
												if (!lastChainId.equals(currentChainId)){
													lastChainId = currentChainId;
													currentChainId = atom.chain_id;
													chainChange = true;
//													lastResidueName = null;
//													lastResidueId = -1;
												}
												else{
													currentChainId = atom.chain_id;
													chainChange = false;
												}
											}
											else{
												lastChainId = currentChainId;
												chainChange = false;
											}


											//there was a residue before
											//this is the first atom of the new residue, so process the
											//last residue for bonds and start up a new one
											HashMap map = ProteinProperties.getBondMapping(lastResidueName);
											if (map != null){
												//go over the atoms and create bonds
												Set keys = residueAtoms.keySet();
												Iterator it = keys.iterator();
												while (it.hasNext()){
													String name = (String)it.next();
													Atom a = (Atom)residueAtoms.get(name);
													
													if (name.equals("N") && lastC != null && !chainChange && dif <= 1){
														//create the bond with the last C
														if (Algebra.distance(lastC.coordinate, a.coordinate) > 2){
//															System.out.println("continue");
															continue;
														}
														Bond b = new Bond(lastC, a);
														b.setOrderOnly(1.0f);
	
														Vector r = (Vector) passComponents.get(
																StructureComponentRegistry.TYPE_BOND );
															if ( r == null )
															{
																r = new Vector( );
																passComponents.put(
																	StructureComponentRegistry.TYPE_BOND, r );
															}
														r.add( b );
													}
													
													Iterator it2 = keys.iterator();
													while (it2.hasNext()){
														String nn = (String)it2.next();
														String label = a.name + ":" + nn;
														if (map.containsKey(label)){
															Bond b = new Bond(a, (Atom)residueAtoms.get(nn));
															float order = ((Float)map.get(label)).floatValue();
															if (order == 1.5f || order == 2.0f){
																a.hybridization = Atom.SP2;
																((Atom)residueAtoms.get(nn)).hybridization = Atom.SP2;
															}
															else if (order == 3.0f){
																a.hybridization = Atom.SP;
																((Atom)residueAtoms.get(nn)).hybridization = Atom.SP;
															}
															
															
															
															b.setOrderOnly(order);
	
															Vector r = (Vector) passComponents.get(
																	StructureComponentRegistry.TYPE_BOND );
																if ( r == null )
																{
																	r = new Vector( );
																	passComponents.put(
																		StructureComponentRegistry.TYPE_BOND, r );
																}
															r.add( b );
														}
													}
													
													//make sure all N's except for Lys are sp2-hybridized
													if (a.element == 7){
														if (a.name.equals("NZ"))continue;//terminal N in lysine
														a.hybridization = Atom.SP2;
													}
												}
											}
											
										}
										else{
											//start of the chain: good place to put a one-time flag
											protein = true;
											dna = false;
											currentChainId = atom.chain_id;
											
										}
										residueAtoms.clear();
										
										lastResidueName = atom.compound;
										lastResidueId = atom.residue_id;
										
									}
									residueAtoms.put(atom.name, atom);
									if (atom.name.equals("C")){
										lastC = currentC;
										currentC = atom;
									}
									
								}
								
								if (DNAProperties.isNucleicAcid(atom.compound)){
									
									dna = true;
									
									//check for lastResidueId that preceded this one
									if (residueAtoms.size() > 0){
										
//										System.out.println("protein = " + protein);
//										System.out.println("dna = " + dna);
										
										if (protein){
											HashMap map = null;
											Set names = residueAtoms.keySet();
//											System.out.println("residueAtoms = " + residueAtoms.size());
											Iterator it = names.iterator();
											while (it.hasNext()){
												
												String name = (String)it.next();
												Atom a = (Atom)residueAtoms.get(name);
												if (a.element == 7){
													if (a.name.equals("NZ"))continue;
													a.hybridization = Atom.SP2;
												}

												if (map == null){
//													System.out.println("assigning map");
													map = ProteinProperties.getBondMapping(a.compound);
													if (map == null){
														//it could be dna
														map = DNAProperties.getBondMapping(a.compound);
													}
													
//													System.out.println("map = " + map);
												}
												
												
												//go over the atoms and create bonds
												
												if (a.name.equals("N")){
													//create the bond with the last C
													if (lastC != null && lastC.residue_id != a.residue_id){
														Bond b = new Bond(lastC, a);
														b.setOrderOnly(1.0f);
//														System.out.println("Created bond with the last C");
														Vector r = (Vector) passComponents.get(
																StructureComponentRegistry.TYPE_BOND );
															if ( r == null )
															{
																r = new Vector( );
																passComponents.put(
																	StructureComponentRegistry.TYPE_BOND, r );
															}
														r.add( b );
													}
												}
															
												Iterator it2 = names.iterator();
												while (it2.hasNext()){
													String nn = (String)it2.next();
													String label = a.name + ":" + nn;
//													System.out.println("map = " + map + ", label = " + label);
//													System.out.println("a = " + a + ", nn = " + residueAtoms.get(nn));
													if (map != null && map.containsKey(label)){
														Bond b = new Bond(a, (Atom)residueAtoms.get(nn));
//														System.out.println("valid map");
														float order = ((Float)map.get(label)).floatValue();
														if (order == 1.5f || order == 2.0f){
															a.hybridization = Atom.SP2;
															((Atom)residueAtoms.get(nn)).hybridization = Atom.SP2;
														}
														else if (order == 3.0f){
															a.hybridization = Atom.SP;
															((Atom)residueAtoms.get(nn)).hybridization = Atom.SP;
														}
														
														b.setOrderOnly(order);
													
														Vector r = (Vector) passComponents.get( StructureComponentRegistry.TYPE_BOND );
														if ( r == null ){
															r = new Vector( );
															passComponents.put( StructureComponentRegistry.TYPE_BOND, r );
														}
														r.add( b );
//														System.out.println("added bond");
													}
												}
											}
										}
										
										protein = false;
									}

									if (atom.residue_id != lastResidueId){
										//there was a switch of residue or this is only the start of the chain
										if (lastResidueName != null){
											//check if this is a chain change
											//also, this may be needed for gaps in residue numbering
											//to avoid making long bonds that span the gap
											if (lastO != null){
												dif = currentO.residue_id - lastO.residue_id;
											}
											if (lastChainId != null){
												if (!lastChainId.equals(currentChainId)){
													lastChainId = currentChainId;
													currentChainId = atom.chain_id;
													chainChange = true;
//													lastResidueName = null;
//													lastResidueId = -1;
												}
												else{
													currentChainId = atom.chain_id;
													chainChange = false;
												}
											}
											else{
												lastChainId = currentChainId;
												chainChange = false;
											}


											//there was a residue before
											//this is the first atom of the new residue, so process the
											//last residue for bonds and start up a new one
											HashMap map = DNAProperties.getBondMapping(lastResidueName);
											if (map != null){
												//go over the atoms and create bonds
												Set keys = residueAtoms.keySet();
												Iterator it = keys.iterator();
												while (it.hasNext()){
													String name = (String)it.next();
													Atom a = (Atom)residueAtoms.get(name);
													if (a.element == 7){
//														System.out.println("SP2 for " + a.name);
														a.hybridization = Atom.SP2;
													}

													if (name.equals("P") && lastO != null && !chainChange && dif <= 1){
														//create the bond with the last C
														Bond b = new Bond(lastO, a);
														b.setOrderOnly(1.0f);
	
														Vector r = (Vector) passComponents.get(
																StructureComponentRegistry.TYPE_BOND );
															if ( r == null )
															{
																r = new Vector( );
																passComponents.put(
																	StructureComponentRegistry.TYPE_BOND, r );
															}
														r.add( b );
													}
													
													Iterator it2 = keys.iterator();
													while (it2.hasNext()){
														String nn = (String)it2.next();
														String label = a.name + ":" + nn;
														if (map.containsKey(label)){
															Bond b = new Bond(a, (Atom)residueAtoms.get(nn));
															float order = ((Float)map.get(label)).floatValue();
															if (order == 1.5f || order == 2.0f){
																a.hybridization = Atom.SP2;
																((Atom)residueAtoms.get(nn)).hybridization = Atom.SP2;
															}
															//higher bond orders are not checked, since this is a standard set of amino
															//acids that don't have triple bonds
															
															b.setOrderOnly(order);
	
															Vector r = (Vector) passComponents.get(
																	StructureComponentRegistry.TYPE_BOND );
																if ( r == null )
																{
																	r = new Vector( );
																	passComponents.put(
																		StructureComponentRegistry.TYPE_BOND, r );
																}
															r.add( b );
														}
													}
													
												}
											}
											
											
										}
										else{
											//start of the chain: good place to put a one-time flag
											protein = false;
											dna = true;
											currentChainId = atom.chain_id;
										}
										residueAtoms.clear();
										
										lastResidueName = atom.compound;
										lastResidueId = atom.residue_id;
										
									}
									residueAtoms.put(atom.name, atom);
									if (atom.name.equals("O3*") || atom.name.equals("O3'")){
										lastO = currentO;
										currentO = atom;
									}
									
								}
							
							}
							catch (Exception ex){
								ex.printStackTrace();
							}
							
//							System.out.println("records = " + records.size());
							// Reset linePos to the start of the line buffer.
							linePos = 0;
							continue;
						}
	
						if (
								(line[0] == 'C') &&
								(line[1] == 'O') &&
								(line[2] == 'N') &&
								(line[3] == 'E') &&
								(line[4] == 'C') &&
								(line[5] == 'T')
							){
							
							if (conect == null){
								//initialize the conect map
								conect = new int[hetatmCount][hetatmCount];
								for (int j = 0; j < hetatmCount; j++){
									for (int k = 0; k < hetatmCount; k++){
										conect[j][k] = 0;
									}
								}
							}
							//parse out the bonding record
							try{
								String record = new String(line);
//								System.out.println("record = " + record);
								int n = record.indexOf("\n");
								int r = record.indexOf("\r");
								String template = null;
								if (n > 0 && r > 0){
									template = record.substring(6, Math.min(n, r));
								}
								else if (r > 0){
									template = record.substring(6, r);
								}
								else{
									template = record.substring(6, n);
								}
								StringTokenizer tok = new StringTokenizer(template, " ", false);
								int root = -1;//index of the root atom
								
								while (tok.hasMoreTokens()){
									try{
										int nn = Integer.parseInt(tok.nextToken());
										//the number may also refer to an atom in regular ATOM records
										//when it's a disulfide bond or a bond between a heteroatom and a
										//protein atom.
										//if the number is not in the lookup hash, find the atom
										//in the records read so far
										/**
										 * TODO
										 */
/*										if (!numberOrder.containsKey(new Integer(nn))){
											if (root == -1){
												//this is the root atom and it's a part of the protein
												ww
											}
										}
*/										if (root == -1){
											//this is the root atom
											root = ((Integer)numberOrder.get(new Integer(nn))).intValue();
											continue;
										}
										conect[root][((Integer)numberOrder.get(new Integer(nn))).intValue()]++;
									}
									catch (Exception ex){
//										System.out.println("ex = " + ex);
//										ex.printStackTrace();
										//handles misformatted conect records. if the first atom is fine,
										//the subsequent errors only result in a loss of one connection
										continue;
									}

								}
							}
							catch (Exception ex){
								System.err.println("Erreur : " + ex);
								ex.printStackTrace();
							}
							
							
							linePos = 0;
							
						}
						//
						// HELIX record
						//
						if (
							(line[0] == 'H') &&
							(line[1] == 'E') &&
							(line[2] == 'L') &&
							(line[3] == 'I') &&
							(line[4] == 'X')
						)
						{
							Helix helix = new Helix( );
	
							helix.name = (new String( line, 0, 6 )).trim();
							helix.start_compound = (new String( line, 15, 3 )).trim();
							helix.start_chain = (new String( line, 19, 1 )).trim();
							if ( helix.start_chain == null ) helix.start_chain = "A";
							helix.start_residue = Integer.parseInt( (new String( line, 21, 4 )).trim() );
							helix.end_compound = (new String( line, 27, 3 )).trim();
							helix.end_chain = (new String( line, 31, 1 )).trim();
							if ( helix.end_chain == null ) helix.end_chain = "A";
							helix.end_residue = Integer.parseInt( (new String( line, 33, 4 )).trim() );
							helix.righthand = true; // JLM DEBUG
	
							Vector records = (Vector) passComponents.get(
								StructureComponentRegistry.TYPE_HELIX );
							if ( records == null )
							{
								records = new Vector( );
								passComponents.put(
									StructureComponentRegistry.TYPE_HELIX, records );
							}
							records.add( helix );
	
							// Reset linePos to the start of the line buffer.
							linePos = 0;
							continue;
						}
	
						//
						// TURN record
						//
						if (
							(line[0] == 'T') &&
							(line[1] == 'U') &&
							(line[2] == 'R') &&
							(line[3] == 'N')
						)
						{
							Turn turn = new Turn( );
	
							turn.name = (new String( line, 0, 6 )).trim();
							turn.start_compound = (new String( line, 15, 3 )).trim();
							turn.start_chain = (new String( line, 19, 1 )).trim();
							if ( turn.start_chain == null ) turn.start_chain = "A";
							turn.start_residue = Integer.parseInt( (new String( line, 20, 4 )).trim() );
							turn.end_compound = (new String( line, 26, 3 )).trim();
							turn.end_chain = (new String( line, 30, 1 )).trim();
							if ( turn.end_chain == null ) turn.end_chain = "A";
							turn.end_residue = Integer.parseInt( (new String( line, 31, 4 )).trim() );
	
							Vector records = (Vector) passComponents.get(
								StructureComponentRegistry.TYPE_TURN );
							if ( records == null )
							{
								records = new Vector( );
								passComponents.put(
									StructureComponentRegistry.TYPE_TURN, records );
							}
							records.add( turn );
	
							// Reset linePos to the start of the line buffer.
							linePos = 0;
							continue;
						}
	
						//
						// STRAND (SHEET) record
						//
						if (
							(line[0] == 'S') &&
							(line[1] == 'H') &&
							(line[2] == 'E') &&
							(line[3] == 'E') &&
							(line[4] == 'T')
						)
						{
							Strand strand = new Strand( );
	
							// Each individual SHEET record is really a single STRAND.
							// And it takes multiple SHEET records in order to
							// fully specify a sheet structure.
							strand.name = (new String( line, 0, 6 )).trim();
							strand.start_compound = (new String( line, 17, 3 )).trim();
							strand.start_chain = (new String( line, 21, 1 )).trim();
							if ( strand.start_chain == null ) strand.start_chain = "A";
							strand.start_residue = Integer.parseInt( (new String( line, 22, 4 )).trim() );
							strand.end_compound = (new String( line, 28, 3 )).trim();
							strand.end_chain = (new String( line, 32, 1 )).trim();
							if ( strand.end_chain == null ) strand.end_chain = "A";
							strand.end_residue = Integer.parseInt( (new String( line, 33, 4 )).trim() );
	
							Vector records = (Vector) passComponents.get(
								StructureComponentRegistry.TYPE_STRAND );
							if ( records == null )
							{
								records = new Vector( );
								passComponents.put(
									StructureComponentRegistry.TYPE_STRAND, records );
							}
							records.add( strand );
	
							// Reset linePos to the start of the line buffer.
							linePos = 0;
							continue;
						}

						// Reset linePos to the start of the line buffer.
						linePos = 0;
					}
					else
					{
						// Continue building the line.
						linePos++;
						if ( linePos >= line.length )
						{
//							System.err.println( "line buffer overflow!" );
							return null;
						}
					}
				}
			}
			
			//create bonds in the last residue that never was reached (for proteins and DNA only)
			if (residueAtoms.size() > 0){
				
//				System.out.println("protein = " + protein);
//				System.out.println("dna = " + dna);
				
				if (protein){
					HashMap map = null;
					Set names = residueAtoms.keySet();
					Iterator it = names.iterator();
					while (it.hasNext()){
						
						String name = (String)it.next();
						Atom a = (Atom)residueAtoms.get(name);
						if (map == null){
//							System.out.println("assigning map");
							map = ProteinProperties.getBondMapping(a.compound);
							if (map == null){
								//it could be dna
								map = DNAProperties.getBondMapping(a.compound);
							}
							
//							System.out.println("map = " + map);
						}
						//go over the atoms and create bonds
						
						if (a.name.equals("N")){
							//create the bond with the last C
							if (lastC != null && lastC.residue_id != a.residue_id){
								Bond b = new Bond(lastC, a);
								b.setOrderOnly(1.0f);
	
								Vector r = (Vector) passComponents.get(
										StructureComponentRegistry.TYPE_BOND );
									if ( r == null )
									{
										r = new Vector( );
										passComponents.put(
											StructureComponentRegistry.TYPE_BOND, r );
									}
								r.add( b );
							}
						}
									
						Iterator it2 = names.iterator();
						while (it2.hasNext()){
							String nn = (String)it2.next();
							String label = a.name + ":" + nn;
//							System.out.println("map = " + map + ", label = " + label);
//							System.out.println("a = " + a + ", nn = " + residueAtoms.get(nn));
							if (map != null && map.containsKey(label)){
								Bond b = new Bond(a, (Atom)residueAtoms.get(nn));
//								System.out.println("valid map");
								float order = ((Float)map.get(label)).floatValue();
								if (order == 1.5f || order == 2.0f){
									a.hybridization = Atom.SP2;
									((Atom)residueAtoms.get(nn)).hybridization = Atom.SP2;
								}
								else if (order == 3.0f){
									a.hybridization = Atom.SP;
									((Atom)residueAtoms.get(nn)).hybridization = Atom.SP;
								}
								
								b.setOrderOnly(order);
							
								Vector r = (Vector) passComponents.get( StructureComponentRegistry.TYPE_BOND );
								if ( r == null ){
									r = new Vector( );
									passComponents.put( StructureComponentRegistry.TYPE_BOND, r );
								}
								r.add( b );
//								System.out.println("added bond");
							}
						}
						if (a.element == 7){
							if (a.name.equals("NZ"))continue;
							a.hybridization = Atom.SP2;
						}
					}
					
					protein = false;
				}
			
			
				if (dna){
					HashMap map = null;
					Set names = residueAtoms.keySet();
					Iterator it = names.iterator();
					while (it.hasNext()){
						
						String name = (String)it.next();
						Atom a = (Atom)residueAtoms.get(name);
						if (map == null){
							map = DNAProperties.getBondMapping(a.compound);
							if (map == null){
								map = ProteinProperties.getBondMapping(a.compound);
							}
						}
						//go over the atoms and create bonds
						
						if (a.name.equals("P")){
							//create the bond with the last O
							if (lastO != null && lastO.residue_id != a.residue_id){
								Bond b = new Bond(lastO, a);
								b.setOrderOnly(1.0f);
	
								Vector r = (Vector) passComponents.get(
										StructureComponentRegistry.TYPE_BOND );
									if ( r == null )
									{
										r = new Vector( );
										passComponents.put(
											StructureComponentRegistry.TYPE_BOND, r );
									}
								r.add( b );
							}
						}
									
						Iterator it2 = names.iterator();
						while (it2.hasNext()){
							String nn = (String)it2.next();
							String label = a.name + ":" + nn;
							if (map != null && map.containsKey(label)){
								Bond b = new Bond(a, (Atom)residueAtoms.get(nn));
								float order = ((Float)map.get(label)).floatValue();
								if (order == 1.5f || order == 2.0f){
									a.hybridization = Atom.SP2;
									((Atom)residueAtoms.get(nn)).hybridization = Atom.SP2;
								}
								
								b.setOrderOnly(order);
							
								Vector r = (Vector) passComponents.get(
										StructureComponentRegistry.TYPE_BOND );
									if ( r == null )
									{
										r = new Vector( );
										passComponents.put(
											StructureComponentRegistry.TYPE_BOND, r );
									}
								r.add( b );
							}
						}
						
						//make sure all N's except for Lys are sp2-hybridized
						if (a.element == 7){
							a.hybridization = Atom.SP2;
						}
					}
					
					dna = false;
				}
				
			}
			
			
			//go through the conect records, if any, and create bonds
			if (hetatmCount > 0 && conect != null){
				for (int i = 0; i < hetatmCount; i++){
					for (int j = i; j < hetatmCount; j++){
						if (i == j) continue;
						//get the real atom numbers
						int order = conect[i][j];
						if (order == 0) continue;
						
						Atom a1 = (Atom)hetatms.get(i);
						Atom a2 = (Atom)hetatms.get(j);
						
						//create a bond
						Bond bond = new Bond(a1, a2);
						if (order == 1){
							bond.setOrderOnly(1.0f);//the default is already SP3 and atoms with another value should not be touched
						}
						else if (order == 2){
							bond.setOrderOnly(2.0f);
							a1.hybridization = Atom.SP2;
							a2.hybridization = Atom.SP2;
						}
						else if (order == 3){
							bond.setOrderOnly(3.0f);
							a1.hybridization = Atom.SP;
							a2.hybridization = Atom.SP;
						}
						
						Vector records = (Vector) passComponents.get(
								StructureComponentRegistry.TYPE_BOND );
							if ( records == null )
							{
								records = new Vector( );
								passComponents.put(
									StructureComponentRegistry.TYPE_BOND, records );
							}
						records.add( bond );

					}
				}
			}
			
		}
		catch (NumberFormatException ex){
			ex.printStackTrace();
			if (Status.enabled) Status.progress( 1.0f, null );
			throw ex;
		}
		catch( IOException e )
		{
			e.printStackTrace();
			// Progress is done.
			if (Status.enabled) Status.progress( 1.0f, null );
			return null;
		}
		catch (Exception ex){
			ex.printStackTrace();
			return null;
		}
		
		if (currentName != null){
			//combine urlString and currentName - for NMR models
			urlString += ":" + currentName;
		}
		
		//IMPORTANT: if chains are renamed, also scan conformational records to make sure they have correct chain ids
		//use their start and end residues
		
		//check ter records
		if (ter.size() > 0){
			//check for chain ids
			Vector records = (Vector) passComponents.get(StructureComponentRegistry.TYPE_ATOM );
			if ( records != null ){
				for (int i = 0; i < records.size(); i++){
					Atom a = (Atom)records.get(i);
					if (a.chain_id.equals("_")){
						//check the last atom to see if there was an id
						if (i >= 1){
							Atom last = (Atom)records.get(i-1);
							if (ter.contains(last)){
								//we have just started a new chain
								String ci = last.chain_id;
								
								//it can't be "_", since we have just passed over it, so figure out what letter it has
								//it can be just an increment for the replacements, or a real id that was there from the start
								int lastIndex = -1;
								for (int j = 0; j < chainIdList.length; j++){
									if (ci.equals(chainIdList[j])){
										chainIdCounter = j+1;
										break;
									}
								}
							}
							
							//it may be necessary to switch to the next chain
						}
						
						//reset it to a lettered chain id
						a.chain_id = chainIdList[chainIdCounter];
					}
					
				}
			}
			
		}
		
		//
		// Create the Structure object
		//
		Structure structure = new Structure()
		{
			// A hashtable of vectors where
			// each hash KEY is the StructureComponent type String.
			// each hash VALUE is a Vector of StructureComponent objects.
			protected Hashtable structureComponents = null;

			// To free up the global state for another load call.
			private String localUrlString;

			// public Structure()  Anonymous inner class constructor.
			{
				structureComponents = passComponents;
				passComponents = null;

				localUrlString = urlString;
				urlString = null;
				
			}

			public String getUrlString( ){
				return localUrlString;
			}

			
			public int getStructureComponentCount( String scType ){
				Vector records = (Vector) structureComponents.get( scType );
				if ( records == null )
					return 0;
				else
					return records.size( );
			}

			private void getStructureComponentByIndex( int index,
				StructureComponent structureComponent )
				throws IndexOutOfBoundsException, IllegalArgumentException
			{
				String type = structureComponent.getStructureComponentType();
				if ( type == null )
					throw new IllegalArgumentException( "type inconnu" + type );

				Vector records = (Vector) structureComponents.get( type );
				if ( records == null )
					throw new IllegalArgumentException( "pas d\'enregistrement du type " + type );

				StructureComponent cachedComponent =
					(StructureComponent) records.elementAt( index );

				cachedComponent.structure = this;
				structureComponent.structure = this;
				structureComponent.copy( cachedComponent );
            }

			public StructureComponent getStructureComponentByIndex( String type,
				int index )
				throws IndexOutOfBoundsException, IllegalArgumentException
			{
				Vector records = (Vector) structureComponents.get( type );
				if ( records == null )
					throw new IllegalArgumentException( "pas d\'enregistrement du type " + type );

				StructureComponent structureComponent =
					(StructureComponent) records.elementAt( index );
				structureComponent.structure = this;

				return structureComponent;
            }
		};

		// Progress is done.
		Status.progress( 1.0f, null );

		return structure;
	}
	
	public int getDataType(){
		return DATA_TYPE;
	}
		
}

