//Copyright (c) 2000-2003  San Diego Supercomputer Center (SDSC),
//a facility operated by the University of California, San Diego (UCSD)
//
//Users and possessors of this source code are hereby granted a
//nonexclusive, royalty-free copyright and design patent license to
//use this code in individual software.  License is not granted for
//commercial resale, in whole or in part, without prior written
//permission from SDSC.  This source is provided "AS IS" without express
//or implied warranty of any kind.
//
//For further information, please see:  http://mbt.sdsc.edu
//
package edu.sdsc.sirius.io;


import edu.sdsc.mbt.*;
import edu.sdsc.mbt.util.*;

import java.util.*;
import java.io.*;
import java.net.*;

public class PqrEntryLoader implements FileEntryLoader, java.io.Serializable {

	protected String urlString = null;
	private long expectedInputBytes = 1;
	
	private final int DATA_TYPE = Entry.DATA_STRUCTURE;

	// A hashtable of vectors where
	// each hash KEY is the StructureComponent type String.
	// each hash VALUE is a Vector of StructureComponent objects.
	// NOTE: This protected variable is used only to enable each new
	// Structure object to access the newly loaded data. The Structure
	// constructor stores a copy of the object so this variable can
	// be re-used for the next load request. This is not thread-safe
	// but there doesn't seem to be any other way to enable an
	// anonymous class to be handed state (aside from parameters).
	protected Hashtable passComponents = null;


	//
	// StructureLoader interface methods
	//
	/**
	 * Returns name of the loader.
	 */
	public String getLoaderName(){
		return new String("Format PQR");
	}

	/**
	 * Loads the Entry by file name.
	 */
	public Entry load( String name, int number ){
	
		StructureEntry entry = new StructureEntry();
		
		Structure structure = null;//future structureObject derived from the PDB file
		try
		{	
			File file = new File( name );
			structure = loadStructure( file, null );
		}
		catch( NullPointerException e )
		{
			try
			{	
				URL url = new URL( name );
				structure = loadStructure( url, null );
				entry.setSourceUrl(url);
			}
			catch( MalformedURLException e2 )
			{
			}
		}
		if ( structure == null ){
			return null;
		}
		
		urlString = name;
		
		//modify the file name to get rid of the extension and use this string as the entry name
		String shortName = name;
		if (name.endsWith(".pqr") || name.endsWith(".PQR")){
			shortName = name.substring(0, name.length()-4);
		}
		
		if(shortName.indexOf(File.separator) > -1){
			int index = shortName.lastIndexOf(File.separator);
			String temp = shortName.substring(index+1);
			shortName = temp;
		}
		
		entry.setStructure(structure);
		
		entry.setName(shortName);
		
		return entry;
	}
	
	
	
	/**
	 * Loads the Entry from a File object.
	 */
	public Entry load(File file, int number){
	
		StructureEntry entry = new StructureEntry();
		
		String name = file.toString();
//		urlString = name;
		
		Structure structure = null;
		try {	
			structure = loadStructure( file, null );
			entry.setSourceUrl(file.toURL());
		}
		catch( Exception e ){}
		
		if ( structure == null ){
			return null;
		}
		
		//modify the file name to get rid of the extension and use this string as the entry name
		String shortName = name;
		if (name.endsWith(".pqr") || name.endsWith(".PQR")){
			shortName = name.substring(0, name.length()-4);
		}

		if(shortName.indexOf(File.separator) > -1){
			int index = shortName.lastIndexOf(File.separator);
			String temp = shortName.substring(index+1);
			shortName = temp;
		}

		entry.setStructure(structure);
		entry.setName(shortName);
		
		return entry;
	
	
	}
	

	
	
	/**
	 * Loads the Entry from a URL object.
	 */
	public Entry load(URL url, int number){
	
		StructureEntry entry = new StructureEntry();
		
		String name = url.toString();
		
		Structure structure = loadStructure( url, null );

		if ( structure == null ){
			return null;
		}
		
		entry.setSourceUrl(url);
		
		//modify the file name to get rid of the extension and use this string as the entry name
		String shortName = name;
		if (name.endsWith(".pqr") || name.endsWith(".PQR")){
			shortName = name.substring(0, name.length()-4);
		}

		if(shortName.indexOf(File.separator) > -1){
			int index = shortName.lastIndexOf(File.separator);
			String temp = shortName.substring(index+1);
			shortName = temp;
		}

		entry.setStructure(structure);
		entry.setName(shortName);
		
		return entry;
	}

	
	/**
	 * Returns true if the loader is capable of loading the structure,
	 * or false otherwise. This enables higher-level code to be able
	 * to build a context sensative menu of only the loaders that can
	 * load a given structure name.
	 */
	public boolean canLoad( String name )
	{
		// System.err.println( "PdbStructureLoader.canLoad(String)" );
		if ( name.endsWith( ".pqr" ) || name.endsWith(".PQR") )
			return true;

		return false;
	}

	//
	// FileStructureLoader interface methods
	//
	
	/**
	 * Returns a reference to a Structure read from the given File object.
	 */
	public Structure loadStructure( File file, String identificator )
	{
		// System.err.println( "PdbStructureLoader.load(File)" );
		
//		System.out.println("loadStructure with id = " + identificator);

		if ( ! canLoad( file ) ) return null;
		if ( ! file.exists() ) return null;
		if ( ! file.canRead() ) return null;

		expectedInputBytes = file.length( );
		if ( expectedInputBytes <= 0 ) return null;

		try
		{
			FileInputStream fileInputStream = new FileInputStream( file );
			if ( fileInputStream == null ) return null;
			
			//filename was being reduced to just the name itself here and in SequenceViewerPanel
//			urlString = file.toURL().toExternalForm();
			
			String tempUrlString = null;
			if (identificator == null){
				tempUrlString = file.toURL().toExternalForm();
				
				//process the name to get rid of the full path and keep
				//only the file name without the extension for internal reference purposes
				int index = tempUrlString.lastIndexOf("/") + 1;
				String tUrlString = tempUrlString.substring(index);
			
				index = tUrlString.lastIndexOf(".");
				urlString = tUrlString.substring(0,index);
			}
			else{
				urlString = identificator;
//				System.out.println("tempUrlString = " + tempUrlString);
			}
			
			
			BufferedInputStream bufferedInputStream = new BufferedInputStream( fileInputStream );
			
			return loadStructure( bufferedInputStream );
		}
		catch ( FileNotFoundException e )
		{
			return null;
		}
		catch ( MalformedURLException e )
		{
			return null;
		}
		catch ( SecurityException e )
		{
			return null;
		}
	}


	/**
	 * Returns true if the Entry can be read from the given File object.
	 */
	public boolean canLoad( File file )
	{
		if ( ! file.exists() ) return false;

		try
		{
			// String path = file.getCanonicalPath( );
			String path = file.toURL().toExternalForm();
			return canLoad( path );
		}
		catch( IOException e )
		{
			return false;
		}
	}


	/**
	 * Returns a reference to a Structure read from the given URL object.
	 */
	public Structure loadStructure( URL url, String identificator )
	{

		if ( ! canLoad( url ) ) return null;

		try
		{
			URLConnection urlConnection = url.openConnection( );
			expectedInputBytes = urlConnection.getContentLength( );
			if ( expectedInputBytes <= 0 ) return null;
			InputStream inputStream = urlConnection.getInputStream( );
			
			if (identificator != null){
				urlString = identificator;
			}

			BufferedInputStream bufferedInputStream = null;
			if ( inputStream != null ) urlString = url.toExternalForm( );
			bufferedInputStream = new BufferedInputStream( inputStream );
			
			return loadStructure( bufferedInputStream );
		}
		catch( IOException e )
		{
			return null;
		}
	}


	/**
	 * Returns true if the Entry can be read from the given URL object.
	 */
	public boolean canLoad( URL url )
	{
		return canLoad( url.toExternalForm( ) );
	}

	//
	// PqrStructureLoader methods
	//

	/**
	 * Returns a reference to a Structure read from the given (uncompressed)
	 * InputStream.
	 */
	public Structure loadStructure( BufferedInputStream bufferedInputStream )
	{
		// System.err.println( "PdbStructureLoader.load(BufferedInputStream)" );
		if ( bufferedInputStream == null ) return null;

		passComponents = new Hashtable( );
		long expectedBytes = expectedInputBytes;
		// System.err.println( "PdbStructureLoader.load: expectedBytes = " + expectedBytes );

		byte buf[] = new byte[ 2048 ]; // raw read data buffer
		byte line[] = new byte[ 100 ]; // accumulated line buffer

		int expectedReads = (int) ((float) expectedBytes / (float) buf.length);
		// System.err.println( "PdbStructureLoader.load: expectedReads = " + expectedReads );
		int onePercent = (int) (((float) expectedBytes / (float) buf.length) / 100.f);
		if ( onePercent <= 0 ) onePercent = 1;
		// System.err.println( "PdbStructureLoader.load: onePercent = " + onePercent );
		float percentDone = 0.0f;

		Status.progress( percentDone, "Chargement " + urlString );
		
		
		int linePos = 0;
		int bytesRead = 0;
		int lastRead = 0;
		int lines = 0;
		int readCount = 0;
		try
		{
			while ( bufferedInputStream.available() > 0 )
			{
				// Re-fill the raw data buffer.
				try
				{
					lastRead = bufferedInputStream.read( buf, 0, buf.length );
				}
				catch( IOException e )
				{
//					System.err.println( "PqrStructureLoader.load: " + e );
					return null;
				}
				bytesRead += lastRead;

				readCount++;
				if ( readCount % onePercent == 0 )
				{
					percentDone = (float) bytesRead / (float) expectedBytes;
					Status.progress( percentDone, "Chargement " + urlString );
				}
	
				// Process the buffer
				for ( int bufPos=0; bufPos<lastRead; bufPos++ )
				{
					// Copy a byte from the input buffer into the line buffer
					line[linePos] = buf[bufPos];
					if ( buf[bufPos] == '\n' )
					{
						// line[0-linePos] now has a complete line.
						lines++;
	
						// Parse the line buffer
	
						//
						// ATOM and HETATM records
						//
						boolean isAtom = false;
						if (
							(line[0] == 'A') &&
							(line[1] == 'T') &&
							(line[2] == 'O') &&
							(line[3] == 'M')
						)
							isAtom = true;
						else if (
							(line[0] == 'H') &&
							(line[1] == 'E') &&
							(line[2] == 'T') &&
							(line[3] == 'A') &&
							(line[4] == 'T') &&
							(line[5] == 'M')
						)
							isAtom = true;
						if ( isAtom )
						{
							// PDB File Atom Record offsets as documented by
							// http://www.rcsb.org/pdb/docs/format/pdbguide2.2/part_62.html
 							// 1 -  6   RecordName
 							// 7 - 11   serial
 							// 12       -
							// 13 - 16  name
							// 17       altLoc
							// 18 - 20  resName
							// 21       -
							// 22       chainID
							// 23 - 26  resSeq
							// 27       iCode
							// 28 - 30  -
							// 31 - 38  x
							// 39 - 46  y
							// 47 - 54  z
							// 55 - 60  occupancy
							// 61 - 66  tempFactor
							// 67 - 72  -
							// 73 - 76  segID
							// 77 - 78  element
							// 79 - 80  charge
							// NOTE: In this application, we need to subtract 1 from
							// each index in order to match the 0-based array offsets.
	
							Atom atom = new Atom( );
							String str = null;

							atom.number =  Integer.parseInt( (new String( line, 6, 5 )).trim() );

							atom.name = (new String( line, 12, 4 )).trim();
							atom.setElement(atom.name.substring(0,1));

							atom.compound = (new String( line, 17, 3 )).trim();
							
							atom.chain_id = (new String( line, 21, 1 )).trim();
							if ( atom.chain_id == null ){
								atom.chain_id = "_";
								
								//try to get the residue number from differently placed fields
								try{
									atom.residue_id =	Integer.parseInt( (new String( line, 22, 4 )).trim() );
								}
								catch (Exception ex){
									atom.residue_id =	Integer.parseInt( (new String( line, 23, 4 )).trim() );
								}
							}
							else if ( atom.chain_id.equals( "" ) ){
								atom.chain_id = "_";
								
								//try to get the residue number from differently placed fields
								try{
									atom.residue_id =	Integer.parseInt( (new String( line, 22, 4 )).trim() );
								}
								catch (Exception ex){
									try{
										atom.residue_id =	Integer.parseInt( (new String( line, 23, 4 )).trim() );
									}
									catch (Exception e){
										atom.residue_id = 1;
									}
								}
							}
							else{
								
								//try to get the residue number from differently placed fields
								try{
									atom.residue_id =	Integer.parseInt( (new String( line, 22, 5 )).trim() );
								}
								catch (Exception ex){
									atom.residue_id =	Integer.parseInt( (new String( line, 23, 5 )).trim() );
								}
							}
							
							atom.coordinate = new double[3];
							atom.coordinate[0] =
								Double.parseDouble( (new String( line, 30, 8 )).trim() );
							atom.coordinate[1] =
								Double.parseDouble( (new String( line, 38, 8 )).trim() );
							atom.coordinate[2] =
								Double.parseDouble( (new String( line, 46, 8 )).trim() );

							str = (new String( line, 54, 6 )).trim();
							if ( str.length() > 0 )
								atom.partialCharge = Float.parseFloat( str );

							str = (new String( line, 63, 6 )).trim();
							if ( str.length() == 0 ){
//								atom.radius = 1.0f;
							}
							else{
								try{
//									atom.radius = Float.parseFloat( str );
								}
								catch (NumberFormatException e){
//									atom.radius = 0.0f;
								}
							}
							Vector records = (Vector) passComponents.get(
								StructureComponentRegistry.TYPE_ATOM );
							if ( records == null )
							{
								records = new Vector( );
								passComponents.put(
									StructureComponentRegistry.TYPE_ATOM, records );
							}
							records.add( atom );
	
							// Reset linePos to the start of the line buffer.
							linePos = 0;
							continue;
						}
	
						//
						// HELIX record
						//
						if (
							(line[0] == 'H') &&
							(line[1] == 'E') &&
							(line[2] == 'L') &&
							(line[3] == 'I') &&
							(line[4] == 'X')
						)
						{
							Helix helix = new Helix( );
	
							helix.name = (new String( line, 0, 6 )).trim();
							helix.start_compound = (new String( line, 15, 3 )).trim();
							helix.start_chain = (new String( line, 19, 1 )).trim();
							if ( helix.start_chain == null ) helix.start_chain = "A";
							helix.start_residue = Integer.parseInt( (new String( line, 21, 4 )).trim() );
							helix.end_compound = (new String( line, 27, 3 )).trim();
							helix.end_chain = (new String( line, 31, 1 )).trim();
							if ( helix.end_chain == null ) helix.end_chain = "A";
							helix.end_residue = Integer.parseInt( (new String( line, 33, 4 )).trim() );
							helix.righthand = true; // JLM DEBUG
	
							Vector records = (Vector) passComponents.get(
								StructureComponentRegistry.TYPE_HELIX );
							if ( records == null )
							{
								records = new Vector( );
								passComponents.put(
									StructureComponentRegistry.TYPE_HELIX, records );
							}
							records.add( helix );
	
							// Reset linePos to the start of the line buffer.
							linePos = 0;
							continue;
						}
	
						//
						// TURN record
						//
						if (
							(line[0] == 'T') &&
							(line[1] == 'U') &&
							(line[2] == 'R') &&
							(line[3] == 'N')
						)
						{
							Turn turn = new Turn( );
	
							turn.name = (new String( line, 0, 6 )).trim();
							turn.start_compound = (new String( line, 15, 3 )).trim();
							turn.start_chain = (new String( line, 19, 1 )).trim();
							if ( turn.start_chain == null ) turn.start_chain = "A";
							turn.start_residue = Integer.parseInt( (new String( line, 20, 4 )).trim() );
							turn.end_compound = (new String( line, 26, 3 )).trim();
							turn.end_chain = (new String( line, 30, 1 )).trim();
							if ( turn.end_chain == null ) turn.end_chain = "A";
							turn.end_residue = Integer.parseInt( (new String( line, 31, 4 )).trim() );
	
							Vector records = (Vector) passComponents.get(
								StructureComponentRegistry.TYPE_TURN );
							if ( records == null )
							{
								records = new Vector( );
								passComponents.put(
									StructureComponentRegistry.TYPE_TURN, records );
							}
							records.add( turn );
	
							// Reset linePos to the start of the line buffer.
							linePos = 0;
							continue;
						}
	
						//
						// STRAND (SHEET) record
						//
						if (
							(line[0] == 'S') &&
							(line[1] == 'H') &&
							(line[2] == 'E') &&
							(line[3] == 'E') &&
							(line[4] == 'T')
						)
						{
							Strand strand = new Strand( );
	
							// Each individual SHEET record is really a single STRAND.
							// And it takes multiple SHEET records in order to
							// fully specify a sheet structure.
							strand.name = (new String( line, 0, 6 )).trim();
							strand.start_compound = (new String( line, 17, 3 )).trim();
							strand.start_chain = (new String( line, 21, 1 )).trim();
							if ( strand.start_chain == null ) strand.start_chain = "A";
							strand.start_residue = Integer.parseInt( (new String( line, 22, 4 )).trim() );
							strand.end_compound = (new String( line, 28, 3 )).trim();
							strand.end_chain = (new String( line, 32, 1 )).trim();
							if ( strand.end_chain == null ) strand.end_chain = "A";
							strand.end_residue = Integer.parseInt( (new String( line, 33, 4 )).trim() );
	
							Vector records = (Vector) passComponents.get(
								StructureComponentRegistry.TYPE_STRAND );
							if ( records == null )
							{
								records = new Vector( );
								passComponents.put(
									StructureComponentRegistry.TYPE_STRAND, records );
							}
							records.add( strand );
	
							// Reset linePos to the start of the line buffer.
							linePos = 0;
							continue;
						}

						// Reset linePos to the start of the line buffer.
						linePos = 0;
					}
					else
					{
						// Continue building the line.
						linePos++;
						if ( linePos >= line.length )
						{
							System.err.println( "dpassement de capacit du tampon de ligne !" );
							return null;
						}
					}
				}
			}
		}
		catch( IOException e )
		{
			System.err.println( "PdbStructureLoader.load: " + e );
			return null;
		}

		//
		// Create the Structure object
		//
		Structure structure = new Structure()
		{
			// A hashtable of vectors where
			// each hash KEY is the StructureComponent type String.
			// each hash VALUE is a Vector of StructureComponent objects.
			protected Hashtable structureComponents = null;

			// To free up the global state for another load call.
			private String localUrlString;

			// public Structure()  Anonymous inner class constructor.
			{
				structureComponents = passComponents;
				passComponents = null;

				localUrlString = urlString;
				urlString = null;
				
			}

			public String getUrlString( ){
				return localUrlString;
			}

			
			public int getStructureComponentCount( String scType ){
				Vector records = (Vector) structureComponents.get( scType );
				if ( records == null )
					return 0;
				else
					return records.size( );
			}

			private void getStructureComponentByIndex( int index,
				StructureComponent structureComponent )
				throws IndexOutOfBoundsException, IllegalArgumentException
			{
				String type = structureComponent.getStructureComponentType();
				if ( type == null )
					throw new IllegalArgumentException( "type inconnu" + type );

				Vector records = (Vector) structureComponents.get( type );
				if ( records == null )
					throw new IllegalArgumentException( "pas d\'enregistrement du type " + type );

				StructureComponent cachedComponent =
					(StructureComponent) records.elementAt( index );

				cachedComponent.structure = this;
				structureComponent.structure = this;
				structureComponent.copy( cachedComponent );
            }

			public StructureComponent getStructureComponentByIndex( String type,
				int index )
				throws IndexOutOfBoundsException, IllegalArgumentException
			{
				Vector records = (Vector) structureComponents.get( type );
				if ( records == null )
					throw new IllegalArgumentException( "pas d\'enregistrement du type " + type );

				StructureComponent structureComponent =
					(StructureComponent) records.elementAt( index );
				structureComponent.structure = this;

				return structureComponent;
            }
		};

		// Progress is done.
		Status.progress( 1.0f, null );

		return structure;
	}
	
	public int getDataType(){
		return DATA_TYPE;
	}
		
}

